import logging
import shlex
import subprocess
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from . import board
from .model import (
    BoardModel,
    CompilerModel,
    CoreModel,
    MetadataModel,
    TripleModel,
    VersionString,
)
from .model.common import OS, Environment, Machine, Toolchain

if TYPE_CHECKING:
    import automate.context


class Compiler(object):
    """Represents an unconfigured generic compiler"""

    def __init__(
        self,
        context: "automate.context.AutomateContext",
        compiler: CompilerModel,
    ):
        self.model = compiler
        self.context = context
        self.logger = logging.getLogger(__name__)

    @property
    def triples(self) -> List[TripleModel]:
        """ List of supported triples """
        return self.model.triples

    @property
    def version(self) -> VersionString:
        """Compiler version"""
        return self.model.version

    @property
    def multiarch(self) -> bool:
        """Wether this compiler supports multiarch rootfs"""
        return self.model.multiarch

    @property
    def basedir(self) -> Path:
        return Path(self.model.basedir)

    @property
    def bin_path(self) -> Path:
        """Installation path of compiler tools"""
        return Path(self.model.basedir) / "bin"

    @property
    def prefix(self):
        """Prefix of compiler tools eg: arm-linux-gnueabihf"""
        return self.model.prefix

    @property
    def cc(self) -> str:
        """Binary name of C compiler"""
        return self.model.prefix + self.model.cc

    @property
    def cxx(self) -> str:
        """Binary name of C++ compiler"""
        return self.model.prefix + self.model.cxx

    @property
    def asm(self) -> str:
        """Binary name of assembler"""
        return self.model.prefix + self.model.asm

    @property
    def ld(self) -> str:
        """Binary name of Linker"""
        return self.model.prefix + self.model.ld

    @property
    def ar(self) -> str:
        """Binary name of Archiver"""
        return self.model.prefix + self.model.ar

    @property
    def toolchain(self) -> Toolchain:
        """Toolchain family of the compiler eg. LLVM or GCC"""
        return self.model.toolchain

    @property
    def id(self) -> str:
        """Unique identifier of compiler in metadata"""
        self.logger.warning("WARNING: this method has been deprecated")
        return self.model.name

    @property
    def name(self) -> str:
        """Unique identifier of compiler in metadata"""
        return self.model.name

    @property
    def runtime(self) -> str:
        return self.model.runtime


class CrossCompiler(Compiler):
    """Represents a Compiler with board specific configuration"""

    def __init__(
        self,
        context: "automate.context.AutomateContext",
        compiler: CompilerModel,
        board: "board.Board",
    ) -> None:
        super(CrossCompiler, self).__init__(context, compiler)

        self.logger = logging.getLogger(__name__)
        self.board = board

        self.logger.debug(
            "Getting compiler {} for {}".format(compiler.name, board.name)
        )
        self.check_multiarch = True
        self.core = 0
        self._flags: List[str] = ["-O2"]
        self._cflags: List[str] = []
        self._cxxflags: List[str] = []
        self._ldflags: List[str] = []
        self._libs: List[str] = []
        self._enable_sysroot = True
        self._isa_opt = True
        self._uarch_opt = True

    def configure_extend(
        self,
        flags: Optional[str] = None,
        cflags: Optional[str] = None,
        cxxflags: Optional[str] = None,
        ldflags: Optional[str] = None,
        libs: Optional[str] = None,
    ) -> None:
        """ Extend compiler flags

        For arguments see configure
        """

        if flags is not None:
            self._flags.extend(shlex.split(flags))

        if cflags is not None:
            self._cflags.extend(shlex.split(cflags))

        if cxxflags is not None:
            self._cxxflags.extend(shlex.split(cxxflags))

        if ldflags is not None:
            self._ldflags.extend(shlex.split(ldflags))

        if libs is not None:
            self._libs.extend(shlex.split(libs))

    def configure(
        self,
        flags: Optional[str] = None,
        cflags: Optional[str] = None,
        cxxflags: Optional[str] = None,
        ldflags: Optional[str] = None,
        libs: Optional[str] = None,
        uarch_opt=True,
        isa_opt=True,
        enable_sysroot=True,
    ) -> None:
        """ Set compiler options

    # Arguments
    flags: Basic flags (used for compilation and linkags)
    cflags: flags for C compiler
    cxxflags: flags for C++ compiler
    ldflags: flags for linker (linker is assumed to call C/C++-Compiler)
    libs: Additional libraries to link (some builds might need -lrt)
    uarch_opt: Enable microarchitecture specific optimizations
    isa_opt: Enable isa specific optimizations
    enable_sysroot: Link and build with --sysroot set to a dump of the boards root file system, linkers and cmake will link to libraries installed on the board
    """

        if flags is not None:
            self._flags = shlex.split(flags)

        if cflags is not None:
            self._cflags = shlex.split(cflags)

        if cxxflags is not None:
            self._cxxflags = shlex.split(cxxflags)

        if ldflags is not None:
            self._ldflags = shlex.split(ldflags)

        if libs is not None:
            self._libs = shlex.split(libs)

        self._isa_opt = isa_opt
        self._uarch_opt = uarch_opt
        self._enable_sysroot = enable_sysroot

    @property
    def gcc_toolchain(self) -> Union[None, "CrossCompiler"]:
        """GCC toolchain to use for LLVM based cross compilers

        The gcc toolchain is used to provide the linker,
        and the runtime libraries libgcc and libstdc++
        """
        if self.toolchain == Toolchain.LLVM:
            return self.board.compiler(toolchain=Toolchain.GCC)
        return None

    @property
    def os(self) -> OS:
        """OS part of compiler target triple"""
        os = self.board.os.triple.os
        assert isinstance(os, OS)
        return os

    @property
    def machine(self) -> Machine:
        """Machine part of compiler target triple"""
        m = self.board.os.triple.machine
        assert isinstance(m, Machine)
        return m

    @property
    def environment(self) -> Environment:
        """Environment part of compiler target triple"""
        e = self.board.os.triple.environment
        assert isinstance(e, Environment)
        return e

    @property
    def isa_flags(self) -> str:
        """Default isa specific flags for this board if enabled"""

        if not self._isa_opt:
            return ""

        isa = self.board.cores[self.core].isa
        return self.model.isa_map.get(isa, "")

    @property
    def uarch_flags(self) -> str:
        """Default microarchitecture specific flags for this board if enabled"""
        if not self._uarch_opt:
            return ""

        uarch = self.board.cores[self.core].uarch
        ret = self.model.uarch_map.get(uarch, "")
        return str(ret)

    @property
    def uarch_or_isa_flags(self) -> str:
        """Default flags machine specific flags for this board if enabled"""
        core = self.board.cores[self.core]
        flag = self.uarch_flags
        if not flag:
            flag = self.isa_flags

        return flag

    @property
    def features(self) -> List[str]:
        features = self.board.cores[self.core].extensions
        ret = []
        for k, v in self.model.feature_map.items():
            if k in features:
                ret.append(f"+{v}")
            else:
                ret.append(f"-{v}")
        return ret

    @property
    def sysroot(self) -> Union[Path, str]:
        """Sysroot flag for this compiler and board"""
        if not Path(self.board.os.sysroot).exists():
            self.logger.warning(
                "Could not find sysroot {} using generic sysroot".format(
                    self.board.os.sysroot
                )
            )
            return ""
        return Path(self.board.os.sysroot)

    @property
    def valid(self) -> bool:
        """Boolean flag wether this compiler is expected to generate working executables
        """
        os_triple = (
            self.board.os.triple.os,
            self.board.os.triple.machine,
            self.board.os.triple.environment,
        )

        for ct in self.model.triples:
            if os_triple == (ct.os, ct.machine, ct.environment):
                if self.check_multiarch and self.board.os.multiarch:
                    if not self.model.multiarch:
                        return False
                return True

        return False

    @property
    def base_flags(self) -> str:
        """basic flags shared between  C/C++ compiler and Linker
        """
        flags = []

        if self.toolchain == Toolchain.LLVM:
            assert self.gcc_toolchain is not None
            flags.append(f"--gcc-toolchain={self.gcc_toolchain.basedir}")
            flags.append(f"-ccc-gcc-name {self.gcc_toolchain.cc}")
            os_triple = self.board.os.triple
            flags.append(
                f"-target {os_triple.machine.value}-{os_triple.os.value}-{os_triple.environment.value}"
            )

        if self.uarch_or_isa_flags:
            flags.append(self.uarch_or_isa_flags)

        if self.sysroot:
            flags.append("--sysroot={}".format(self.sysroot))
        else:
            # Current llvm installations do not find a correct libc
            # so we call die underlying gcc to find its sysroot
            # and set it explicitly
            if self.toolchain == Toolchain.LLVM:
                assert self.gcc_toolchain is not None
                command = f'"{self.gcc_toolchain.bin_path}/{self.gcc_toolchain.cc}" -print-sysroot'

                p = subprocess.Popen(
                    shlex.split(command),
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    stdin=subprocess.PIPE,
                )

                stdout, stderr = p.communicate()
                stdout_dec = stdout.decode("utf-8")
                sysroot = stdout_dec.strip()
                flags.append(f"--sysroot={sysroot}")

        if self._flags:
            flags.extend(self._flags)

        return " ".join(flags)

    @property
    def cflags(self) -> str:
        """CFLAGS for this compiler
        """
        flags = []

        base_flags = self.base_flags
        if base_flags:
            flags.append(base_flags)

        if self._cflags:
            flags.extend(self._cflags)

        return " ".join(flags)

    @property
    def cxxflags(self) -> str:
        """CXXFLAGS for this compiler
        """

        flags = []

        base_flags = self.base_flags
        if base_flags:
            flags.append(base_flags)

        if self.toolchain == Toolchain.LLVM:
            flags.append("--stdlib=libstdc++")

        if self._cxxflags:
            flags.extend(self._cxxflags)

        return " ".join(flags)

    @property
    def ldflags(self):
        """LDFLAGS for this compiler

        These flags are appended to the linker commandline before object files
        """
        flags = []

        base_flags = self.base_flags
        if self.base_flags:
            flags.append(self.base_flags)

        if self.toolchain == Toolchain.LLVM:
            flags.append("--rtlib=libgcc")
            flags.append("--unwindlib=libgcc")
            flags.append("--stdlib=libstdc++")

        if self._ldflags:
            flags.extend(self._ldflags)

        return " ".join(flags)

    @property
    def libs(self) -> str:
        """LIBFLAGS for this compiler

        These flags are appended to the linker driver commandline after the objectfiles
        Currently not used
        """

        libs: List[str] = []

        if self._libs:
            libs.extend(self._libs)

        return " ".join(libs)

    def _system_includes(self) -> List[str]:
        """ TODO: remove """
        if self.toolchain == Toolchain.LLVM:
            assert self.gcc_toolchain is not None
            return self.gcc_toolchain._system_includes()
        includes = []

        command = f'"{self.bin_path}/{self.cc}" {self.cflags} -E -Wp,-v -'

        p = subprocess.Popen(
            shlex.split(command),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
        )

        stdout, stderr = p.communicate("".encode("utf-8"))
        stderr_dec = stderr.decode("utf-8")
        for line in stderr_dec.split("\n"):
            line = line.strip()
            if line.startswith("/"):
                includes.append(line)
        return includes

    @property
    def default_builddir(self) -> Path:
        """ The default build directory for this cross compiler / board combinarion
            For now this is just "<cwd>/builds/<board_id>"
        """
        return Path("builds") / str(self.board.name)

    def __deepcopy__(self, memo) -> "CrossCompiler":
        copy = CrossCompiler(self.context, self.model, self.board)

        copy.check_multiarch = deepcopy(self.check_multiarch, memo)
        copy.core = deepcopy(self.core, memo)
        copy._flags = deepcopy(self._flags, memo)
        copy._cflags = deepcopy(self._cflags, memo)
        copy._cxxflags = deepcopy(self._cxxflags, memo)
        copy._ldflags = deepcopy(self._ldflags, memo)
        copy._libs = deepcopy(self._libs, memo)
        copy._enable_sysroot = deepcopy(self._enable_sysroot, memo)
        copy._isa_opt = deepcopy(self._isa_opt, memo)
        copy._uarch_opt = deepcopy(self._uarch_opt, memo)

        return copy



from enum import Enum
from typing import Any, Dict, List

from pydantic import BaseModel, Field

from .common import *
from .model_base import *


class TripleModel(DataModelBase):
    """Description of a target triple"""

    machine: Machine
    vendor: str = ""
    os: OS
    environment: Environment


class CompilerModel(LoadedModelBase):
    name: str
    triples: List[TripleModel] = Field(
        ..., description="List of supported target triples"
    )
    toolchain: Toolchain
    version: VersionString
    basedir: str
    runtime: str = ""
    cc: str
    cxx: str
    asm: str
    ld: str
    ar: str
    isa_map: Dict[str, str]
    uarch_map: Dict[str, str]
    feature_map: Dict[str, str] = Field(default_factory=dict)
    description: str = ""
    prefix: str = ""
    postfix: str = ""
    multiarch: bool = Field(
        False,
        description="Flag to indicate that this compiler supports builds with multiarch sysroots",
    )



from typing import Dict, List, Optional

from .model_base import DataModelBase, LoadedModelBase


class UserModel(DataModelBase):
    name: str
    mail: str
    public_keys: List[str]


class UsersModel(LoadedModelBase):
    users: Dict[str, UserModel]



import os
from pathlib import Path


def runtime_dir() -> Path:
    env_dir = os.getenv("XDG_RUNTIME_DIR")
    if env_dir is None:
        dir = Path("~/.automate").expanduser()
    else:
        dir = Path(env_dir)

    dir = dir / "automate"
    dir.mkdir(exist_ok=True, parents=True)

    return dir



import logging
from pathlib import Path, PosixPath
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

from ruamel.yaml import YAML

from ..utils.network import rsync
from .builder import BaseBuilder

if TYPE_CHECKING:
    import automate.compiler


class MakefileBuilder(BaseBuilder):
    def configure(
        self,
        cross_compiler: "Optional[automate.compiler.CrossCompiler]" = None,
        srcdir: Union[Path, str] = "",
        prefix: Union[Path, str] = "",
        extra_flags: Optional[Dict[str, str]] = None,
        override_flags: Optional[Dict[str, str]] = None,
    ):
        """ Configure a makefile build

            1. Copy source directory to build directory
            2. Record build variables in build_directory/buildvars.yml
        """

        super(MakefileBuilder, self).configure(
            cross_compiler=cross_compiler,
            srcdir=srcdir,
            prefix=prefix,
            extra_flags=extra_flags,
            override_flags=override_flags,
        )

        cross_compiler = self.cross_compiler

        self._mkbuilddir()

        if Path.is_file(Path(f"{self.srcdir}/.automateignore")):
            self.context.run(f"rsync -ar --exclude \'.automateignore\' --exclude-from={self.srcdir}/.automateignore --delete {self.srcdir} {self.builddir}")
        else:
            self.context.run(f"rsync -ar --exclude \'.automateignore\' --delete {self.srcdir} {self.builddir}")

        buildvars: Dict[str, Any] = {}
        buildvars["CC"] = cross_compiler.bin_path / cross_compiler.cc
        buildvars["CXX"] = cross_compiler.bin_path / cross_compiler.cxx
        buildvars["AR"] = cross_compiler.bin_path / cross_compiler.ar
        buildvars["CFLAGS"] = cross_compiler.cflags
        buildvars["CXXFLAGS"] = cross_compiler.cxxflags
        buildvars["LDFLAGS"] = cross_compiler.ldflags
        buildvars["LDLIBS"] = cross_compiler.libs

        self.state.buildvars = buildvars

        self._save_state()

    def build(self, target=""):
        """Run make with default target and set BUILDVARS for board"""

        buildvars = self.state.buildvars

        with self.context.cd(str(self.builddir / self.srcdir.name)):
            self.context.run(
                f"make -j{self._num_build_cpus()} {target} CC=\"{buildvars['CC']}\" CXX=\"{buildvars['CXX']}\" AR=\"{buildvars['AR']}\" CFLAGS=\"{buildvars['CFLAGS']}\" CXXFLAGS=\"{buildvars['CXXFLAGS']}\" LDFLAGS=\"{buildvars['LDFLAGS']}\" LDLIBS=\"{buildvars['LDLIBS']}\""
            )

    def install(self):
        """Do nothing"""
        logging.warning("Install does nothing with make builder")

    def deploy(self, connection=None, delete=False):
        """Deploy package on board

           Just copies build_directory/srcdir_name to the rundir

           # Arguments

           delete: if true delete remove non existant files from install prefix on the board
        """
        if connection == None:
            with self.board.connect() as con:
                self.deploy_internal(con, delete)
        else:
            self.deploy_internal(connection, delete)

    def deploy_internal(self, con, delete=False):
        """Deploy package on board (see deploy())

           # Arguments

           con: Connection object

           delete: see deploy()
        """
        with con.cd(str(self.board.rundir)):
            con.run(f"mkdir -p {self.srcdir.name}")
        rsync(
            con,
            source=str(self.builddir / self.srcdir.name) + "/",
            target=str(self.board.rundir / self.srcdir.name),
            delete=delete,
            rsync_opts="-l",
        )




from typing import Dict, Tuple

arm_cpus: Dict[int, str] = {
    0x810: "arm810",
    0x920: "arm920",
    0x922: "arm922t",
    0x926: "arm926ej-s",
    0x940: "arm940t",
    0x946: "arm946e-s",
    0x966: "arm966e-s",
    0xA20: "arm1020e",
    0xA22: "arm1022e",
    0xA26: "arm1026ej-s",
    0xB02: "arm11-mpcore",
    0xB36: "arm1136j-s",
    0xB56: "arm1156t2-s",
    0xB76: "arm1176jz-s",
    0xC05: "cortex-a5",
    0xC07: "cortex-a7",
    0xC08: "cortex-a8",
    0xC09: "cortex-a9",
    0xC0D: "cortex-a17",
    0xC0F: "cortex-a15",
    0xC0E: "cortex-a17",
    0xC14: "cortex-r4",
    0xC15: "cortex-r5",
    0xC17: "cortex-r7",
    0xC18: "cortex-r8",
    0xC20: "cortex-m0",
    0xC21: "cortex-m1",
    0xC23: "cortex-m3",
    0xC24: "cortex-m4",
    0xC27: "cortex-m7",
    0xC60: "cortex-m0plus",
    0xD01: "cortex-a32",
    0xD03: "cortex-a53",
    0xD04: "cortex-a35",
    0xD05: "cortex-a55",
    0xD07: "cortex-a57",
    0xD08: "cortex-a72",
    0xD09: "cortex-a73",
    0xD0A: "cortex-a75",
    0xD0B: "cortex-a76",
    0xD0C: "neoverse-n1",
    0xD13: "cortex-r52",
    0xD20: "cortex-m23",
    0xD21: "cortex-m33",
    0xD4A: "neoverse-e1",
}

broadcom_cpus: Dict[int, str] = {
    0x00F: "brahma-b15",
    0x100: "brahma-b53",
    0x516: "thunderx2t99",
}

dec_cpus: Dict[int, str] = {0xA10: "sa110", 0xA11: "sa1100"}

cavium_cpus: Dict[int, str] = {
    0x0A0: "thunderx",
    0x0A1: "thunderxt88",
    0x0A2: "thunderxt81",
    0x0A3: "thunderxt83",
    0x0AF: "thunderx2t99",
}

apm_cpu: Dict[int, str] = {0x000: "xgene1"}

qualcom_cpus: Dict[int, str] = {
    0x00F: "scorpion",
    0x02D: "scorpion",
    0x04D: "krait",
    0x06F: "krait",
    0x201: "kryo",
    0x205: "kryo",
    0x211: "kryo",
    0xC00: "falkor",
    0xC01: "saphira",
}

samsung_cpus: Dict[int, str] = {0x001: "exynos_m1"}

nvidia_cpus: Dict[int, str] = {
    0x000: "denver",
    0x003: "denver2",
    0x004: "carmel",
}

marvell_cpus: Dict[int, str] = {
    0x131: "feroceon_88fr131",
    0x581: "pj4-pj4b",
    0x584: "pj4b-mp",
}

faraday_cpus: Dict[int, str] = {0x526: "fa526", 0x626: "fa626"}

intel_cpus: Dict[int, str] = {
    0x200: "i80200",
    0x210: "pxa250a",
    0x212: "pxa210a",
    0x242: "i80321-400",
    0x243: "i80321-600",
    0x290: "pxa250b_pxa26x",
    0x292: "pxa210b",
    0x2C2: "i80321_400_b0",
    0x2C3: "i80321_600_b0",
    0x2D0: "pxa250c/pxa255/pxa26x",
    0x2D2: "pxa210c",
    0x411: "pxa27x",
    0x41C: "ipx425_533",
    0x41D: "ipx425_400",
    0x41F: "ipx425_266",
    0x682: "pxa32x",
    0x683: "pxa930/pxa935",
    0x688: "pxa30x",
    0x689: "pxa31x",
    0xB11: "sa1110",
    0xC12: "ipx1200",
}

hisilicon_cpus: Dict[int, str] = {0xD01: "tsv110"}


infineon_cpus: Dict[int, str] = {}


implementers: Dict[int, Tuple[str, Dict[int, str]]] = {
    0x41: ("arm", arm_cpus),
    0x42: ("broadcom", broadcom_cpus),
    0x43: ("cavium", cavium_cpus),
    0x44: ("dec", dec_cpus),
    0x48: ("hisilicon", hisilicon_cpus),
    0x49: ("infineon", infineon_cpus),
    0x4E: ("nvidia", nvidia_cpus),
    0x50: ("apm", arm_cpus),
    0x51: ("qualcom", qualcom_cpus),
    0x53: ("qualcom", samsung_cpus),
    0x56: ("marvell", marvell_cpus),
    0x66: ("faraday", faraday_cpus),
    0x69: ("intel", intel_cpus),
}


# TODO: there should be a better way to do this
# Source: https://developer.arm.com/ip-products/processors/cortex-a
uarch_to_isa: Dict[str, str] = {
    "cortex-a12": "armv7-a",
    "cortex-a15": "armv7-a",
    "cortex-a17": "armv7-a",
    "cortex-a32": "armv8-a",
    "cortex-a35": "armv8-a",
    "cortex-a5": "armv7-a",
    "cortex-a53": "armv8-a",
    "cortex-a55": "armv8.2-a",
    "cortex-a57": "armv8-a",
    "cortex-a7": "armv7-a",
    "cortex-a72": "armv8-a",
    "cortex-a73": "armv8.a",
    "cortex-a75": "armv8.2-a",
    "cortex-a76": "armv8.2-a",
    "cortex-a77": "armv8.2-a",
    "cortex-a9": "armv7-a",
    "denver": "armv8-a",
    "denver2": "armv8-a",
    "carmel": "armv8.2-a",
}



from fabric import task

from ..model.common import Toolchain


@task
def configure(
    c,
    board,
    builddir="",
    srcdir="",
    prefix="",
    flags=None,
    cflags=None,
    cxxflags=None,
    ldflags=None,
    libs=None,
    sysroot=True,
    isa=True,
    uarch=True,
    toolchain="gcc",
    compiler_name="",
):  # pragma: no cover

    board = c.board(board)

    toolchain = Toolchain(toolchain) if toolchain else Toolchain.GCC

    cc = board.compiler(toolchain=toolchain, compiler_name=compiler_name)
    cc.configure(
        flags=flags,
        cflags=cflags,
        cxxflags=cxxflags,
        ldflags=ldflags,
        uarch_opt=uarch,
        isa_opt=isa,
        enable_sysroot=sysroot,
        libs=libs,
    )
    builder = board.builder("make", builddir=builddir)

    builder.configure(cc, srcdir=srcdir, prefix=prefix)


@task
def build(c, board, builddir=""):  # pragma: no cover
    board = c.board(board)
    builder = board.builder("make", builddir=builddir)

    builder.build()


@task
def install(c, board, builddir=""):  # pragma: no cover
    board = c.board(board)
    builder = board.builder("make", builddir=builddir)

    builder.install()


@task
def clean(c, board, builddir=""):  # pragma: no cover
    board = c.board(board)
    builder = board.builder("make", builddir=builddir)

    builder.clean()


@task
def deploy(c, board, builddir=""):  # pragma: no cover
    board = c.board(board)
    builder = board.builder("make", builddir=builddir)

    builder.deploy()


__all__ = ["configure", "build", "clean", "install", "deploy"]



from invoke import Call, Executor

from .context import AutomateContext


class AutomateCall(Call):
    def make_context(self, config):
        return AutomateContext(config)


class AutomateExecutor(Executor):
    def expand_calls(self, calls):
        extended_calls = super().expand_calls(calls)

        ret = []
        for call in extended_calls:
            automate_call = call.clone(into=AutomateCall)
            ret.append(automate_call)

        return ret



import getpass
import logging
import shelve
import threading
import time
from collections import namedtuple
from datetime import datetime, timedelta
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union, cast

from .database import Database

if TYPE_CHECKING:
    from automate.board import Board


class KeepLockThread(threading.Thread):
    def __init__(
        self,
        manager,
        board_name,
        current_lease_time,
        lease_time_increase=90,
    ):
        self.manager = manager
        self.board_name = board_name
        self.current_lease_time = current_lease_time
        self.lease_time_increase = max(lease_time_increase, 60)

        self.stop_event = threading.Event()

        logging.debug("Creating keep lock thread")
        super().__init__(daemon=True, name=f"KeepLockThread({board_name})")

    def run(self):

        while True:
            wait_time = max(0, self.current_lease_time - 10)
            logging.debug(
                "Keep lock thread for %s waiting for %d seconds",
                self.board_name,
                wait_time,
            )
            self.stop_event.wait(wait_time)
            if self.stop_event.is_set():
                return
            if self.manager.has_lock(self.board_name):
                logging.info(
                    "Increasing lock time by %d seconds",
                    self.lease_time_increase,
                )
                self.manager.lock(
                    self.board_name, str(self.lease_time_increase)
                )
                self.current_lease_time = self.lease_time_increase
            else:
                self.current_lease_time = 11

    def stop(self):
        logging.debug("Stopping keep lock thread")
        self.stop_event.set()


class LockManagerBase:
    """ Base Class for lock managers """

    def __init__(self):
        pass

    def _do_unlock(self, board_name: str) -> None:
        raise NotImplementedError("_do_unlock is not implemented")

    def _do_trylock(self, board_name: str, lease_time: float) -> bool:
        raise NotImplementedError("_do_trylock is not implemented")

    def _do_haslock(self, board_name) -> bool:
        raise NotImplementedError("_do_haslock is not implemented")

    def _do_islocked(self, board_name) -> bool:
        raise NotImplementedError("_do_haslock is not implemented")

    def _do_lease_time(self, board_name: str) -> timedelta:
        raise NotImplementedError("_do_get_lease_time is not implemented")

    def _do_lock_holder(self, board_name: str) -> str:
        raise NotImplementedError("_do_get_lock_holder is not implemented")

    def _str_to_timedelta(self, inp: str) -> timedelta:
        inp = inp.strip()
        seconds = 0
        if inp[-1] == "d":
            seconds = int(inp[:-1].strip()) * 3600 * 24
        elif inp[-1] == "h":
            seconds = int(inp[:-1].strip()) * 3600
        elif inp[-1] == "m":
            seconds = int(inp[:-1].strip()) * 60
        elif inp[-1] == "s":
            seconds = int(inp[:-1].strip())
        else:
            seconds = int(inp)

        delta = timedelta(seconds=seconds)

        return delta

    def lock(
        self,
        board: Union["Board", str],
        lease_time: Union[timedelta, str] = "1h",
    ) -> None:
        if not self.trylock(board, lease_time):
            logging.warning(
                "Board has already been locked by a different user waiting until board is available"
            )
            while not self.trylock(board, lease_time):
                time.sleep(0.5)

    def unlock(self, board: Union["Board", str]) -> None:

        if isinstance(board, str):
            board_name = board
        else:
            board_name = board.name

        if self._do_haslock(board_name):
            self._do_unlock(board_name)

    def trylock(
        self,
        board: Union["Board", str],
        lease_time: Union[timedelta, str] = "1h",
    ) -> bool:
        if isinstance(board, str):
            board_name = board
        else:
            board_name = board.name

        if isinstance(lease_time, str):
            delta = self._str_to_timedelta(lease_time)
        else:
            delta = lease_time

        if self.has_lock(board_name):
            current_delta = self.lease_time(board_name)
            if current_delta > delta:
                return True

        return self._do_trylock(board_name, delta.total_seconds())

    def has_lock(self, board: Union["Board", str]) -> bool:
        if isinstance(board, str):
            board_name = board
        else:
            board_name = board.name

        return self._do_haslock(board_name)

    def is_locked(self, board: Union["Board", str]) -> bool:
        if isinstance(board, str):
            board_name = board
        else:
            board_name = board.name

        return self._do_islocked(board_name)

    def lease_time(self, board: Union["Board", str]) -> timedelta:
        """ Returns the remaining lock time for the board as a timedelta"""
        if isinstance(board, str):
            board_name = board
        else:
            board_name = board.name
        lease_time = self._do_lease_time(board_name)
        if lease_time.total_seconds() > 0:
            return lease_time
        return timedelta()

    def lock_holder(self, board: Union["Board", str]) -> str:
        if isinstance(board, str):
            board_name = board
        else:
            board_name = board.name
        holder = self._do_lock_holder(board_name)
        lease_time = self.lease_time(board_name)

        if lease_time.total_seconds() > 0:
            return holder
        return ""

    def keep_lock(self, board: Union["Board", str]) -> Optional[KeepLockThread]:
        if isinstance(board, str):
            board_name = board
        else:
            board_name = board.name

        lease_time = 0.0
        if self.has_lock(board_name):
            lease_time = self.lease_time(board_name).total_seconds()

        thread = KeepLockThread(self, board_name, lease_time)
        thread.start()

        return thread


LockEntry = namedtuple("LockEntry", ["user_id", "timestamp"])


class SimpleLockManager(LockManagerBase):
    """Simple lock manager using a gdbm shared file identifying lock holders by the username on the current machine"""

    def __init__(
        self, lockfile: Union[str, Path], user_id: str = "", db=None
    ) -> None:
        super(SimpleLockManager, self).__init__()
        self.lockfile = str(Path(lockfile).absolute())
        self.user_id = user_id
        if not user_id:
            self.user_id = getpass.getuser()

        self.logger = logging.getLogger(__name__)

    def _do_unlock(self, board_name: str) -> None:
        """ releases the lock from board with board_name """
        try:
            with shelve.open(self.lockfile) as lockdb:
                if board_name in lockdb:
                    assert hasattr(lockdb[board_name], "user_id")
                    if (
                        cast(LockEntry, lockdb[board_name]).user_id
                        == self.user_id
                    ):
                        del lockdb[board_name]
        except Exception as e:
            self.logger.error("Exception during board unlock", str(e))

        return None

    def _do_trylock(self, board_name: str, lease_time: float) -> bool:
        """
        checks if board with board_name is locked
        if the desired board is locked it will be checked
            if user owns the lock and if so the lease will be updated
            if the user does not own the lock it will be checked if the lease is still valid
                if the lease is valid the lock will be denied -> False
                if the lease is invalid the lock will be granted -> True
        if the desired board is not locked the the lock will be granted -> True
        """

        delta = timedelta(seconds=lease_time)
        lease_time_absolute = datetime.now() + delta

        try:
            with shelve.open(self.lockfile) as lockdb:
                current_timestamp = datetime.now()
                if board_name in lockdb:
                    current_lock: LockEntry = cast(
                        LockEntry, lockdb[board_name]
                    )
                    if current_lock.user_id != self.user_id:
                        if current_timestamp < current_lock.timestamp:
                            return False
                        else:
                            lockdb[board_name] = LockEntry(
                                self.user_id, lease_time_absolute
                            )
                            return True

                    if current_lock.timestamp < lease_time_absolute:
                        lockdb[board_name] = LockEntry(
                            self.user_id, lease_time_absolute
                        )
                    return True

                lockdb[board_name] = LockEntry(
                    self.user_id, lease_time_absolute
                )
        except Exception as e:
            self.logger.error("Exception during board lock %s", str(e))

        return True

    def _do_haslock(self, board_name: str) -> bool:
        """ checks if user owns the lock for board with board_name """
        try:
            with shelve.open(self.lockfile) as lockdb:
                if board_name in lockdb:
                    current_lock = cast(LockEntry, lockdb[board_name])
                    current_timestamp = datetime.now()

                    if (
                        current_lock.user_id == self.user_id
                        and current_lock.timestamp > current_timestamp
                    ):
                        return True
        except Exception as e:
            self.logger.error("Exception during has_lock %s", str(e))

        return False

    def _do_islocked(self, board_name: str) -> bool:
        """ checks if board with board_name is locked by any user """
        try:
            with shelve.open(self.lockfile) as lockdb:
                if board_name in lockdb:
                    current_lock = cast(LockEntry, lockdb[board_name])
                    current_timestamp = datetime.now()
                    if (
                        current_lock.user_id != self.user_id
                        and current_lock.timestamp > current_timestamp
                    ):
                        return True

        except Exception as e:
            self.logger.error("Exception during board islocked", str(e))

        return False

    def _do_lease_time(self, board_name: str) -> timedelta:
        try:
            with shelve.open(self.lockfile) as lockdb:
                if board_name in lockdb:
                    current_lock = cast(LockEntry, lockdb[board_name])
                    current_timestamp = datetime.now()
                    lock_timestamp = current_lock.timestamp

                    delta = lock_timestamp - current_timestamp
                    assert isinstance(delta, timedelta)
                    return delta

        except Exception as e:
            self.logger.error("Exception during board islocked: %s", str(e))

        return timedelta()

    def _do_lock_holder(self, board_name: str) -> str:
        try:
            with shelve.open(self.lockfile) as lockdb:
                if board_name in lockdb:
                    current_lock = cast(LockEntry, lockdb[board_name])
                    return str(current_lock.user_id)

        except Exception as e:
            self.logger.error("Exception during board islocked %s", str(e))
        return ""


class DatabaseLockManager(LockManagerBase):
    """ lock manager using the database for distributed locks """

    def __init__(self, database: Database, user_id: str = "") -> None:
        super(DatabaseLockManager, self).__init__()

        self.database = database
        self.user_id = user_id

        if not user_id or user_id == "":
            self.user_id = getpass.getuser()

        self.logger = logging.getLogger(__name__)

    def _do_unlock(self, board_name: str) -> None:
        self.database.unlock(board_name, self.user_id)

    def _do_trylock(self, board_name: str, lease_time: float) -> bool:
        return self.database.trylock(board_name, self.user_id, int(lease_time))

    def _do_haslock(self, board_name: str) -> bool:
        return self.database.haslock(board_name, self.user_id)

    def _do_islocked(self, board_name: str) -> bool:
        return self.database.islocked(board_name, self.user_id)

    def _do_lease_time(self, board_name: str) -> timedelta:
        return self.database.lease_time(board_name)

    def _do_lock_holder(self, board_name: str) -> str:
        return self.database.lock_holder(board_name)



import os
from pathlib import Path
from typing import Any, Dict, Optional

from invoke.config import Config, merge_dicts


class AutomateConfig(Config):
    """Encapsulates ~/.automate.yml"""

    prefix = "automate"
    env_prefix = "AUTOMATE"

    @staticmethod
    def global_defaults() -> Any:
        their_defaults = Config.global_defaults()

        my_defaults = {
            "automate": {
                "metadata": os.path.expanduser("~/.automate/metadata"),
                "metadata_url": "git@es-git.cs.uni-tuebingen.de:/es/timing/schrank/metadata.git",
                "metadata_ref": "main",
                "identity": os.path.expanduser("~/.ssh/id_rsa"),
                "toolroot": os.path.expanduser("~/.automate/tools"),
                "boardroot": os.path.expanduser("~/.automate/boards"),
            },
            "debug": {"level": "INFO"},
            "tasks": {"collection_name": "autofile"},
        }

        return merge_dicts(their_defaults, my_defaults)



from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from ..utils.network import rsync
from .builder import BaseBuilder

if TYPE_CHECKING:
    import automate.board
    import automate.compiler


class CMakeBuilder(BaseBuilder):
    def configure(
        self,
        cross_compiler: "Optional[automate.compiler.CrossCompiler]" = None,
        srcdir: Union[Path, str] = "",
        prefix: Union[Path, str] = "",
        extra_flags: Optional[Dict[str, str]] = None,
        override_flags: Optional[Dict[str, str]] = None,
        cmake_definitions: List[str] = [],
    ):
        """
        Configure cmake based build

        # Arguments 
        cross_compiler: cross compiler for the build, default: use default compiler for the board
        srcdir: directory conaining CMakeLists.txt
        prefix: install prefix on the board
        cmake_definitions: extra cmake definitions 
        """
        super(CMakeBuilder, self).configure(
            cross_compiler=cross_compiler,
            srcdir=srcdir,
            prefix=prefix,
            extra_flags=extra_flags,
            override_flags=override_flags,
        )

        cross_compiler = self.cross_compiler

        self.clean()
        self._mkbuilddir()

        toolchain_file_name = self.builddir / "toolchain.cmake"
        with toolchain_file_name.open("w") as toolchain_file:
            toolchain_file.write("set(CMAKE_SYSTEM_NAME Linux)\n")
            toolchain_file.write(
                "set(CMAKE_SYSTEM_PROCESSOR {})\n".format(
                    cross_compiler.machine.value
                )
            )
            toolchain_file.write("\n")
            toolchain_file.write(
                "set(CMAKE_SYSROOT {})\n".format(cross_compiler.sysroot)
            )
            toolchain_file.write(
                "set(CMAKE_STAGING_PREFIX {})\n".format(
                    self.builddir / "install"
                )
            )
            toolchain_file.write("\n")
            toolchain_file.write(
                "set(CMAKE_C_COMPILER {}/{})\n".format(
                    cross_compiler.bin_path, cross_compiler.cc
                )
            )
            toolchain_file.write(
                "set(CMAKE_CXX_COMPILER {}/{})\n".format(
                    cross_compiler.bin_path, cross_compiler.cxx
                )
            )

            toolchain_file.write(
                "set(CMAKE_AR {}/{})\n".format(
                    cross_compiler.bin_path, cross_compiler.ar
                )
            )
            toolchain_file.write("\n")
            toolchain_file.write(
                "set(CMAKE_FIND_ROOT_PATH_MODE_PROGRAM NEVER)\n"
            )
            toolchain_file.write(
                "set(CMAKE_FIND_ROOT_PATH_MODE_LIBRARY ONLY)\n"
            )
            toolchain_file.write(
                "set(CMAKE_FIND_ROOT_PATH_MODE_INCLUDE ONLY)\n"
            )
            toolchain_file.write(
                "set(CMAKE_FIND_ROOT_PATH_MODE_PACKAGE ONLY)\n"
            )
            toolchain_file.write("\n")

        cache_file_name = self.builddir / "cache.cmake"
        with cache_file_name.open("w") as cache_file:
            cache_file.write("#Compiler options \n")
            cflags = cross_compiler.cflags
            cache_file.write(
                'set(CMAKE_C_FLAGS_DEBUG          "{} -g" CACHE STRING "")\n'.format(
                    cflags
                )
            )
            cache_file.write(
                'set(CMAKE_C_FLAGS_MINSIZEREL     "{} -DNDEBUG" CACHE STRING "")\n'.format(
                    cflags
                )
            )
            cache_file.write(
                'set(CMAKE_C_FLAGS_RELWITHDEBINFO "{} -g -DNDEBUG" CACHE STRING "")\n'.format(
                    cflags
                )
            )
            cache_file.write(
                'set(CMAKE_C_FLAGS_RELEASE        "{} -DNDEBUG" CACHE STRING "")\n'.format(
                    cflags
                )
            )
            cache_file.write("\n")

            cxxflags = cross_compiler.cxxflags
            cache_file.write(
                'set(CMAKE_CXX_FLAGS_DEBUG          "{} -g" CACHE STRING "")\n'.format(
                    cxxflags
                )
            )
            cache_file.write(
                'set(CMAKE_CXX_FLAGS_MINSIZEREL     "{} -DNDEBUG" CACHE STRING "")\n'.format(
                    cxxflags
                )
            )
            cache_file.write(
                'set(CMAKE_CXX_FLAGS_RELWITHDEBINFO "{} -g -DNDEBUG" CACHE STRING "")\n'.format(
                    cxxflags
                )
            )
            cache_file.write(
                'set(CMAKE_CXX_FLAGS_RELEASE        "{} -DNDEBUG" CACHE STRING "")\n'.format(
                    cxxflags
                )
            )

        definitions = " ".join(["-D{}".format(d) for d in cmake_definitions])

        with self.context.cd(str(self.builddir)):
            command = "cmake -DCMAKE_BUILD_WITH_INSTALL_RPATH=ON -DCMAKE_INSTALL_RPATH={2}/lib -DCMAKE_BUILD_TYPE='RELWITHDEBINFO' -DCMAKE_TOOLCHAIN_FILE=toolchain.cmake -C cache.cmake {0} {1}".format(
                self.srcdir, definitions, self.prefix
            )
            self.logger.info("Running cmake: {}".format(command))
            self.context.run(command)

        self._save_state()

    def build(self, c):
        "Run cmake build"
        self._mkbuilddir()
        with self.context.cd(str(self.builddir)):
            self.context.run("cmake --build  .")

    def install(self, c):
        "Run cmake install"
        with self.context.cd(str(self.builddir)):
            self.context.run("cmake  --build . --target install")

    def deploy(self, c, delete=False):
        """ Deploy target on board
        
        # Arguments
        delete: if true clean install prefix before deployment
        """
        print("Rsyncing with prefix", self.prefix)
        with self.board.connect() as con:
            con.run(f"mkdir -p {self.prefix.name}")
            rsync(
                con,
                source=str(self.builddir / "install") + "/",
                target=str(self.prefix),
                delete=delete,
                rsync_opts="-l",
            )



from functools import wraps


# TODO: calling all functions as eg directory(c, '/foo/bar/') (with initial c)
# will probably get old; but what's better?
# - Phrasing them as methods on mixin classes to pile on top of Context (so:
# c.directory('/foo/bar/')) pushes us farther towards a god object antipattern,
# though it seems clearly the most practical, especially if we reimplement
# implicit global context;
# - Have each patchwork module as a property of Context itself might help a
# _little_ bit (so it becomes: c.files.directory('/foo/bar/')) but only a
# little, and it almost too-strongly ties us to the module organization (it's
# much easier to change a 'from x import y' than to change a bunch of
# references to a variable)
# - Standalone objects that encapsulate a Context would help to some degree,
# but only in terms of scaling up to many per-module calls (i.e. filemanager =
# files.Manager(c); filemanager.directory('/foo/bar'); if
# filemanager.exists('blah'): xyz; etc) and seems to scale poorly across
# modules (if you want to use eg 1-2 funcs from each module you've got a
# bunch of objects to keep track of)
# - What other patterns are good?

# TODO: how much of the stuff in here should be in Invoke or Fabric core? For
# now, it's ok to leave here because it's highly experimental, but expect a 2.0
# or 3.0 release to see it move elsewhere once patterns are established.


# TODO: is it possible to both expose the wrapped function's signature to
# Sphinx autodoc AND remain a signature-altering decorator? =/ Using @decorator
# is nice but it is only for signature-PRESERVING decorators. We really want a)
# prevent users from having to phrase runner as 'runner=None', and b) the
# convenience of sudo=True...but that means signature-altering.
# TODO: this may be another spot where we want to reinstate fabric 1's docs
# unwrap function.
def set_runner(f):
    """
    Set 2nd posarg of decorated function to some callable ``runner``.

    The final value of ``runner`` depends on other args given to the decorated
    function (**note:** *not* the decorator itself!) as follows:

    - By default, you can simply ignore the decorated function's ``runner``
      argument entirely, in which case it will end up being set to the ``run``
      method on the first positional arg (expected to be a
      `~invoke.context.Context` and thus, the default value is `Context.run
      <invoke.context.Context.run>`).
    - You can override which method on that object is selected, by handing an
      attribute name string to ``runner_method``.
    - Since the common case for this functionality is to trigger use of
      `~invoke.context.Context.sudo`, there is a convenient shorthand, setting
      ``sudo=True``.
    - Finally, you may give a callable object to ``runner`` directly, in which
      case nothing special really happens (it's largely as if you called the
      function undecorated). This is useful for cases where you're calling one
      decorated function from within another.

    .. note::
        The ``runner_method`` and ``sudo`` kwargs exist only at the decorator
        level, and are not passed into the decorated function.

    .. note::
        If more than one of the above kwargs is given at the same time, only
        one will win, in the following order: ``runner``, then
        ``runner_method``, then ``sudo``.

    As an example, given this example ``function``::

        @set_runner
        def function(c, runner, arg1, arg2=None):
            runner("some command based on arg1 and arg2")

    One may call it without any runner-related arguments, in which case
    ``runner`` ends up being a reference to ``c.run``::

        function(c, "my-arg1", arg2="my-arg2")

    Or one may specify ``sudo`` to trigger use of ``c.sudo``::

        function(c, "my-arg1", arg2="my-arg2", sudo=True)

    If one is using a custom Context subclass with other command runner
    methods, one may give ``runner_method`` explicitly::

        class AdminContext(Context):
            def run_admin(self, *args, **kwargs):
                kwargs['user'] = 'admin'
                return self.sudo(*args, **kwargs)

        function(AdminContext(), "my-arg1", runner_method='run_admin')

    Finally, to reiterate, you can always give ``runner`` directly to avoid any
    special processing (though be careful not to get mixed up - if this runner
    isn't actually a method on the ``c`` context object, debugging could be
    frustrating!)::

        function(c, "my-arg1", runner=some_existing_runner_object)
    """

    @wraps(f)
    def inner(*args, **kwargs):
        args = list(args)
        # Pop all useful kwargs (either to prevent clash with real ones, or to
        # remove ones not intended for wrapped function)
        runner = kwargs.pop("runner", None)
        sudo = kwargs.pop("sudo", False)
        runner_method = kwargs.pop("runner_method", None)
        # Figure out what gets applied and potentially overwrite runner
        if not runner:
            method = runner_method
            if not method:
                method = "sudo" if sudo else "run"
            runner = getattr(args[0], method)
        args.insert(1, runner)
        return f(*args, **kwargs)

    return inner



import logging
import multiprocessing
import os
import shutil
from copy import deepcopy
from pathlib import Path
from string import Template
from typing import TYPE_CHECKING, Dict, Optional, Union

from ruamel.yaml import YAML

from ..compiler import CrossCompiler
from ..utils import untar
from ..utils.kernel import KernelConfigBuilder, KernelData
from ..utils.network import rsync
from ..utils.uboot import build_ubimage

if TYPE_CHECKING:
    import automate.board
    import automate.context


class BuilderState:
    builddir: Path  # Build Directory
    srcdir: Path  # Source Direcotry
    prefix: Path  # Install Prefix
    kernel: Optional[Dict[str, str]] = None
    buildvars: Optional[Dict[str, str]] = None


class BaseBuilder(object):
    def __init__(
        self,
        context: "automate.context.AutomateContext",
        board: "automate.board.Board",
        builddir: Union[Path, str] = "",
    ):
        self.logger = logging.getLogger(__name__)
        self.board = board
        self.context = context

        if builddir:
            builddir = Path(builddir)
        else:
            builddir = Path("builds") / board.id

        assert isinstance(builddir, Path)
        if not self._load_state(builddir):
            self.state = BuilderState()
            self.state.builddir = builddir
            self.state.srcdir = Path(".")
            self.state.prefix = Path(self.board.rundir) / self.srcdir.name

            self.state.srcdir = self.state.srcdir.absolute()
            self.state.builddir = self.state.builddir.absolute()

    @property
    def builddir(self) -> Path:
        return self.state.builddir

    @property
    def srcdir(self) -> Path:
        return self.state.srcdir

    @property
    def prefix(self) -> Path:
        return self.state.prefix

    def _num_build_cpus(self) -> int:
        """ Return number of local cpus for build"""
        return multiprocessing.cpu_count()

    def _save_state(self) -> None:
        self._mkbuilddir()
        yaml = YAML(typ="unsafe")
        state_file = self.builddir / ".builder_state.yml"
        with state_file.open("w") as state_file_obj:
            yaml.dump(self.state, state_file_obj)

    def _load_state(self, builddir: Path) -> bool:
        state_file = builddir / ".builder_state.yml"
        if state_file.exists():
            yaml = YAML(typ="unsafe")
            with state_file.open() as state_file_obj:
                self.state = yaml.load(state_file_obj)

            return True

        return False

    def configure(
        self,
        cross_compiler: Optional[CrossCompiler] = None,
        srcdir: Union[Path, str] = "",
        prefix: Union[Path, str] = "",
        extra_flags: Optional[Dict[str, str]] = None,
        override_flags: Optional[Dict[str, str]] = None,
    ):
        """
        Configure the build
        
        # Arguments
        cross_compler: Cross compiler object to use for the build
        srcdir: source directory
        prefix: install prefix
        extra_flags: dict of flags to append for compilation and linkage
        override_flags: dict of flags to override for compilation and linkage 
        """

        if cross_compiler is not None:
            self.cross_compiler = deepcopy(cross_compiler)
        

        if override_flags is not None:
            self.cross_compiler.configure(**override_flags)

        if extra_flags is not None:
            self.cross_compiler.configure_extend(**extra_flags)

        if srcdir:
            self.state.srcdir = Path(srcdir).absolute()

        if prefix:
            prefix = Path(prefix)
            if prefix.is_absolute():
                self.state.prefix = prefix
            else:
                self.state.prefix = self.board.rundir / prefix

        else:
            self.state.prefix = self.board.rundir / self.state.srcdir.name

    def build(self, *args, **kwargs):
        "Build target code"
        raise NotImplementedError("Build is not implemented")

    def install(self, *args, **kwargs):
        "Installs the build in local directory"
        raise NotImplementedError("Installation is not implemented")

    def deploy(self, *args, **kwargs):
        "Deploys installed binary to the board"
        raise NotImplementedError("Installation is not implemented")

    def clean(self, *args, **kwargs):
        "Removes the builddir"
        if self.builddir != self.srcdir:
            if self.builddir.exists():
                shutil.rmtree(self.builddir)

    def _mkbuilddir(self):
        "Creates the build directory"
        self.builddir.mkdir(parents=True, exist_ok=True)


class SPECBuilder(BaseBuilder):
    # TODO: implement spec Builder
    pass


class AutotoolsBuilder(BaseBuilder):
    # TODO: implement autotools Builder
    pass



import logging

from invoke import task

from ..model import BoardModelDB, BoardModelFS


@task
def init(c):
    """Initializes the database with default tables"""

    try:
        c.database.init()
    except Exception as e:
        logging.error("Database initialization failed: %s", str(e))


@task
def import_board(c, board, keep=False):
    """Upload a board from local collection to database

       # Arguments
       board: name of the board to upload
       keep: keep yml file in metadata dir
    """

    board = c.board(board, expand=False)

    if isinstance(BoardModelDB):
        logging.error("Board board.id has already been imported to database")

    try:
        c.database.insert_board(board)
    except Exception as e:
        logging.error("%s", str(e))
    else:
        if not keep:
            c.run("rm ")


@task
def export_board(c, board, keep=False):
    """Download board data and save as description.yml in metadata dir

       # Arguments
       board: name of the board
       keep: keep database entry of board
    """

    boards = c.database.get_all_boards()
    board_model = None
    for bm in boards:
        if bm.name == board:
            board_model = bm
            break

    if not board_model:
        logging.error("Could not find board %s in database", board_model)
        return -1



from enum import Enum
from typing import Any, Dict, List

from pydantic import BaseModel

from .board import BoardModel
from .compiler import CompilerModel
from .model_base import *


class MetadataModel(DataModelBase):
    compilers: List[CompilerModel]
    boards: List[BoardModel]



import logging
from pathlib import Path

from invoke import Collection, task


@task
def compile(
    c, board, files=[], output="a.out", compiler="", builddir=""
):  # pragma: no cover
    """Compiles multiple source files to a target specific executable

       Output will be placed in: build/{board_id} by default

       -b/--board: taget board
       -f/--files: source code files supports C/C++/Object Code files 
                   .cc, .cxx, .cpp, .C, .c++ are interpreted as C++
                   .o and .obj as Object Code (are only linked in)
                   the rest is given to the C Compiler  
    """

    logging.warning(
        "Using compile tasks to build binaries is probably not safe"
    )
    logging.warning("It might make sense to switch to one of the builders")

    board = c.board(board)
    compiler = board.compiler(compiler)

    logging.info(
        "Compiling {} with compiler {}".format(", ".join(files), compiler.id)
    )

    cc = compiler.bin_path / compiler.cc
    cxx = compiler.bin_path / compiler.cxx

    if not builddir:
        builddir = compiler.default_builddir

    build_path = Path(builddir)
    build_path.mkdir(exist_ok=True, parents=True)

    objs = []
    is_cpp = False

    # Compile
    for f in (Path(f) for f in files):
        if not f.exists():
            raise Exception("{} does not exist".format(f))

        if f.suffix == ".o":
            objs.append(f)
            continue

        obj = build_path / (f.stem + ".o")
        objs.append(obj)

        comp = cc
        flags = compiler.cflags
        if f.suffix in ["cc", "cxx", "cpp", "C", "c++"]:
            comp = cxx
            flags = cxxflags
            is_cpp = True

        cmd_list = [comp, "-c", "-o", obj, flags, f]
        cmd = " ".join((str(i) for i in cmd_list))
        logging.info("COMPILE: {}".format(cmd))
        c.run(cmd)

    # Link
    linker = cc
    if is_cpp:
        linker = cxx

    binary = build_path / output
    cmd_list = [linker, "-o", binary, compiler.ldflags] + objs
    cmd = " ".join((str(i) for i in cmd_list))
    logging.info("LINK: {}".format(cmd))
    c.run(cmd)



import logging
from contextlib import contextmanager
from datetime import timedelta
from os.path import dirname, join
from typing import Any, List

from ..model import (
    BoardModel,
    BoardModelDB,
    CoreModel,
    DocumentationLinkModel,
    KernelImageModel,
    KernelModel,
    OSModel,
    SSHConnectionModel,
    TripleModel,
    UBootModel,
)

try:
    import psycopg2  # type: ignore
    import psycopg2.extras  # type: ignore
    from jinjasql import JinjaSql  # type; ignore
    from psycopg2.pool import ThreadedConnectionPool

    enabled = True
except:
    enabled = False


def database_enabled() -> bool:
    """Returns True if required packages for database connection are found"""
    return enabled


class Database:
    """A Database connection"""

    def __init__(
        self, host: str, port: int, db: str, user: str, password: str
    ) -> None:
        """Init database connection

        # Arguments

        host: hostname or ip address for database connection
        port: port for database connection
        db: name of database
        user: username for connection
        password: password for connection
        """
        self.logger = logging.getLogger(__name__)

        self.host = host
        self.port = port
        self.db = db
        self.user = user
        self.password = password

        try:
            self.connection_pool = ThreadedConnectionPool(
                1, 10, host=host, port=port, user=user, password=password
            )
        except Exception as e:
            self.logger.error(
                "could not connnect to database: 'host=%s port=%d db=%s user=%s password=%s'",
                host,
                port,
                user,
                password,
            )
            self.logger.error(e)

        self.j = JinjaSql(param_style="pyformat")

        self.QUERIES_DIR = join(dirname(__file__), "queries")

        self.all_boards_query = self.__load_query("select_all_boards")
        self.all_cpu_cores_for_board_query = self.__load_query(
            "select_all_cpu_cores_for_board"
        )
        self.os_for_board_query = self.__load_query("select_os_for_board")
        self.all_kernels_for_os_query = self.__load_query(
            "select_all_kernels_for_os"
        )
        self.all_docs_for_board_query = self.__load_query(
            "select_all_docs_for_board"
        )
        self.insert_board_query = self.__load_query("insert_board")
        self.init_database_query = self.__load_query("init_database")

        self.insert_lock_query = self.__load_query("insert_lock")
        self.select_lock_for_board_query = self.__load_query(
            "select_lock_for_board"
        )
        self.delete_lock_for_board_query = self.__load_query(
            "delete_lock_for_board"
        )
        self.transfer_lock_for_board_query = self.__load_query(
            "transfer_lock_for_board"
        )
        self.update_lock_lease_for_board_query = self.__load_query(
            "update_lock_lease_for_board"
        )

    @contextmanager
    def connection(self):
        try:
            conn = self.connection_pool.getconn()
            yield conn
        finally:
            self.connection_pool.putconn(conn)

    @contextmanager
    def cursor(self):
        try:
            conn = self.connection_pool.getconn()
            cursor = conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            yield cursor
        finally:
            self.connection_pool.putconn(conn=conn)

    def __load_query(self, name: str) -> str:
        sql_file_path = self.QUERIES_DIR + "/" + name + ".sql"

        with open(sql_file_path, "r") as sql_file:
            query = sql_file.read()
        return query

    def init(self) -> None:
        """Initialize an database with locks"""
        query = self.init_database_query
        with self.connection() as conn:
            cursor = conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            cursor.execute(query)
            conn.commit()

    def get_all_boards(self) -> List[BoardModelDB]:
        with self.cursor() as cursor:
            cursor.execute(self.all_boards_query)
            boards = cursor.fetchall()

            board_models = []

            for board in boards:
                query, bind_params = self.j.prepare_query(
                    self.all_cpu_cores_for_board_query,
                    {"board_id": board["id"]},
                )
                cursor.execute(query)
                cpu_cores = cursor.fetchall()

                query, bind_params = self.j.prepare_query(
                    self.os_for_board_query, {"board_id": board["id"]}
                )
                cursor.execute(query)
                os = cursor.fetchone()

                query, bind_params = self.j.prepare_query(
                    self.all_docs_for_board_query, {"board_id": board["id"]}
                )
                cursor.execute(query)
                docs = cursor.fetchall()

                query, bind_params = self.j.prepare_query(
                    self.all_kernels_for_os_query, {"board_id": board["id"]}
                )
                cursor.execute(query)
                kernels = cursor.fetchall()

                kernel_models = []

                for kernel in kernels:
                    kernel_image_model = KernelImageModel(
                        **{
                            "build_path": kernel["image_build_path"],
                            "deploy_path": kernel["image_deploy_path"],
                        }
                    )

                    uboot_model = None

                    if (
                        kernel["uboot_loadaddr"] is not None
                        and kernel["uboot_image_name"] is not None
                        and kernel["uboot_dtb_image"] is not None
                    ):
                        uboot_model = UBootModel(
                            **{
                                "loadaddr": kernel["uboot_loadaddr"],
                                "image_name": kernel["uboot_image_name"],
                                "dtb_image": kernel["uboot_dtb_image"],
                            }
                        )

                    kernel_model = KernelModel(
                        **{
                            "name": "",  # A name uniquely identifies a a triple of kernel_configuration/commandline/kernel_source code
                            "description": kernel["description"],
                            "version": kernel["version"],
                            "commandline": kernel["command_line"],
                            "kernel_config": kernel["kernel_config"],
                            "kernel_source": kernel["kernel_source"],
                            "kernel_srcdir": kernel["kernel_srcdir"],
                            "image": kernel_image_model,
                            "uboot": uboot_model,
                            "default": kernel["is_default"],
                        }
                    )

                    kernel_models.append(kernel_model)

                triple_model = TripleModel(
                    **{
                        "machine": os["machine"],
                        "os": os["os"],
                        "environment": os["environment"],
                    }
                )

                os_model = OSModel(
                    **{
                        "triple": triple_model,
                        "distribution": os["distribution"],
                        "release": os["release"],
                        "description": os["description"],
                        "sysroot": os["sysroot"],
                        "rootfs": os["rootfs"],
                        "multiarch": os["multiarch"],
                        "kernels": kernel_models,
                    }
                )

                documentation_link_models = []

                for doc in docs:
                    documentation_link_model = DocumentationLinkModel(**doc)

                    documentation_link_models.append(documentation_link_model)

                cpu_core_models = []

                for cpu_core in cpu_cores:
                    # FIXME: Core Model DB?
                    cpu_core_model = CoreModel(**cpu_core)

                    cpu_core_models.append(cpu_core_model)

                ssh_connection_model = SSHConnectionModel(
                    host=board["hostname"],
                    username=board["ssh_username"],
                    port=board["ssh_port"],
                )

                # FIXME: move to separate table
                board = dict(board)
                del board["ssh_username"]
                del board["ssh_port"]

                # FIXME: add database fields for rundir and board
                board_model = BoardModelDB(  # type: ignore
                    **board,
                    rundir="/home/es/run",
                    board="unknown",
                    doc=documentation_link_models,
                    connections=[ssh_connection_model],
                    cores=cpu_core_models,
                    os=os_model,
                )

                board_models.append(board_model)

        return board_models

    def insert_board(
        self, board_model: BoardModel, additional_data: Any
    ) -> None:
        """Insert a board into database

        # Arguments
        board_model: The BoardModel to insert
        additional_data: to be removed

        TODO: what happens if a board model with the same name already exists.

        """

        cpu_isas = set()
        cpu_implementers = set()
        cpu_uarchs = []
        cpu_extensions = set()

        for cpu_core in board_model.cores:
            cpu_isas.add(cpu_core.isa)
            cpu_implementers.add(cpu_core.vendor)
            cpu_uarchs.append(
                {
                    "id": cpu_core.uarch + cpu_core.vendor,
                    "name": cpu_core.uarch,
                    "vendor": cpu_core.vendor,
                }
            )
            cpu_extensions.update(cpu_core.extensions)

        query, bind_params = self.j.prepare_query(
            self.insert_board_query,
            {
                "soc_name": additional_data["soc_name"],
                "foundry_name": additional_data["foundry_name"],
                "technology": additional_data["technology"],
                "hostname": board_model.hostname,
                "board": board_model,
                "power_connector_name": additional_data["power_connector_name"],
                "voltage": additional_data["voltage"],
                "max_current": additional_data["voltage"],
                "cpu_isas": list(cpu_isas),
                "cpu_implementers": list(cpu_implementers),
                "cpu_uarchs": cpu_uarchs,
                "cpu_part_ids": additional_data["cpu_part_ids"],
            },
        )

        try:
            with self.connection() as conn:
                cursor = conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
                cursor.execute(query)
                conn.commit()
        except Exception as e:
            self.logger.error("ERROR: database import failed")
            self.logger.error(e)

    def _select_lock_for_board(self, board_name: str):
        query, bind_params = self.j.prepare_query(
            self.select_lock_for_board_query, {"board_name": board_name}
        )
        try:
            with self.connection() as conn:
                cursor = conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
                cursor.execute(query)
                conn.commit()
                lock = cursor.fetchone()
        except Exception as e:
            self.logger.error(
                "ERROR: could not access lock for '" + board_name + "'"
            )
            self.logger.error(e)

        return lock

    def unlock(self, board_name: str, user_id: str) -> None:
        query, bind_params = self.j.prepare_query(
            self.delete_lock_for_board_query,
            {"board_name": board_name, "user_id": user_id},
        )
        try:
            with self.connection() as conn:
                cursor = conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
                cursor.execute(query)
                conn.commit()
        except Exception as e:
            self.logger.error("ERROR: unlocking of '" + board_name + "' failed")
            self.logger.error(e)

    def trylock(
        self, board_name: str, user_id: str, lease_duration: int
    ) -> bool:
        # check if board is locked
        lock = self._select_lock_for_board(board_name)

        self.logger.info(
            "Trying to lock board %s for %d seconds", board_name, lease_duration
        )

        # board is locked
        if lock is not None:
            if lock["user_id"] != user_id:
                # board is locked by other user
                if lock["lease"] > lock["current_timestamp"]:
                    return False
                # transfer lock
                else:
                    query, bind_params = self.j.prepare_query(
                        self.transfer_lock_for_board_query,
                        {
                            "board_name": board_name,
                            "user_id": user_id,
                            "lease_duration": lease_duration,
                        },
                    )
                    try:
                        with self.connection() as conn:
                            cursor = conn.cursor(
                                cursor_factory=psycopg2.extras.DictCursor
                            )
                            cursor.execute(query)
                            conn.commit()
                    except Exception as e:
                        self.logger.error(
                            "ERROR: transfer of lock for '"
                            + board_name
                            + "' failed"
                        )
                        self.logger.error(e)

                    return True

            # user has lock and lease will be updated
            else:
                query, bind_params = self.j.prepare_query(
                    self.update_lock_lease_for_board_query,
                    {
                        "board_name": board_name,
                        "user_id": user_id,
                        "lease_duration": lease_duration,
                    },
                )
                try:
                    with self.connection() as conn:
                        cursor = conn.cursor(
                            cursor_factory=psycopg2.extras.DictCursor
                        )
                        cursor.execute(query)
                        conn.commit()
                except Exception as e:
                    self.logger.error(
                        "ERROR: update of lock lease for '"
                        + board_name
                        + "' failed"
                    )
                    self.logger.error(e)

                return True

        # board is not locked -> aquire lock
        else:
            query, bind_params = self.j.prepare_query(
                self.insert_lock_query,
                {
                    "board_name": board_name,
                    "user_id": user_id,
                    "lease_duration": lease_duration,
                },
            )
            try:
                with self.connection() as conn:
                    cursor = conn.cursor(
                        cursor_factory=psycopg2.extras.DictCursor
                    )
                    cursor.execute(query)
                    conn.commit()
            except Exception as e:
                self.logger.error(
                    "ERROR: acquiring lock for '" + board_name + "' failed"
                )
                self.logger.error(e)

            return True

    def haslock(self, board_name: str, user_id: str) -> bool:
        self.logger.debug("Test if %s has lock for %s", user_id, board_name)
        lock = self._select_lock_for_board(board_name)

        self.logger.debug("Current Lock: %s", str(lock))
        if lock is None:
            return False

        if (
            lock["user_id"] == user_id
            and lock["lease"] > lock["current_timestamp"]
        ):
            return True

        return False

    def islocked(self, board_name: str, user_id: str) -> bool:
        self.logger.debug("Test if board is locked %s", board_name)
        lock = self._select_lock_for_board(board_name)
        self.logger.debug("Current Lock: %s", str(lock))
        # no lock exists
        if lock is None:
            return False

        # board is locked
        if (
            lock["lease"] > lock["current_timestamp"]
            and lock["user_id"] != user_id
        ):
            return True

        # lock exists but lease is invalid
        return False

    def lock_holder(self, board_name: str) -> str:
        lock = self._select_lock_for_board(board_name)
        if lock:
            return str(lock["user_id"])

        return ""

    def lease_time(self, board_name: str) -> timedelta:
        lock = self._select_lock_for_board(board_name)
        if lock:
            delta = lock["lease"] - lock["current_timestamp"]
            assert isinstance(delta, timedelta)
            return delta

        return timedelta()



from typing import Dict, Generator, List, Tuple, Union

from .. import board, compiler
from ..model.common import Machine

__all__ = ["KernelConfigBuilder", "KernelData"]


class FilteredKernelOption:
    def __init__(self, option: str) -> None:
        self.option = option

    def filter(self, board: "board.Board") -> bool:
        raise NotImplementedError("filter has not been implemented")


class MachineOption(FilteredKernelOption):
    def __init__(
        self, option: str, machines: List[Union[Machine, str]]
    ) -> None:
        machines = [
            m if isinstance(m, Machine) else Machine(m) for m in machines
        ]

        self.machines = set(machines)

        super(MachineOption, self).__init__(option)

    def filter(self, board: "board.Board") -> bool:
        return board.os.triple.machine in self.machines


class PredefinedNotFoundException(Exception):
    pass


class KernelConfigBuilder:
    def __init__(
        self, board: "board.Board", cross_compiler: "compiler.CrossCompiler"
    ) -> None:
        self.board = board
        self.cross_compiler = cross_compiler
        self._predefined_configs: Dict[str, List[Union[MachineOption, str]]] = {
            "default": [
                # Base Options
                "CONFIG_PROC_KCORE=y",
                "CONFIG_IKCONFIG=y",
                "CONFIG_IKCONFIG_PROC=y",
                "CONFIG_IKHEADERS=m",
                "CONFIG_HW_PERF_EVENTS=y",
                "CONFIG_KEXEC=y",
                "CONFIG_KEXEC_CORE=y",
                "CONFIG_NO_HZ_FULL=y",
                "CONFIG_MAGIC_SYSRQ=y",
                "CONFIG_MAGIC_SYSRQ_DEFAULT_ENABLE=0x1",
                "CONFIG_HOTPLUG_CPU=y",
                # Coresight
                MachineOption("CONFIG_CORESIGHT=y", ["aarch64", "arm"]),
                MachineOption(
                    "CONFIG_CORESIGHT_LINKS_AND_SINKS=y", ["aarch64", "arm"]
                ),
                MachineOption(
                    "CONFIG_CORESIGHT_SOURCE_ETM4X=y", ["aarch64", "arm"]
                ),
                MachineOption("CONFIG_CORESIGHT_STM=y", ["aarch64", "arm"]),
                MachineOption(
                    "CONFIG_CORESIGHT_QCOM_REPLICATOR=y", ["aarch64", "arm"]
                ),
                MachineOption(
                    "CONFIG_CORESIGHT_LINK_AND_SINK_TMC=y", ["aarch64", "arm"]
                ),
                MachineOption(
                    "CONFIG_CORESIGHT_SINK_TPIU=y", ["aarch64", "arm"]
                ),
                MachineOption(
                    "CONFIG_CORESIGHT_SINK_ETBV10=y", ["aarch64", "arm"]
                ),
                MachineOption(
                    "CONFIG_CORESIGHT_SOURCE_ETM3X=y", ["aarch64", "arm"]
                ),
                MachineOption(
                    "CONFIG_CORESIGHT_DYNAMIC_REPLICATOR=y", ["aarch64", "arm"]
                ),
                MachineOption("CONFIG_CORESIGHT_CATU=y", ["aarch64", "arm"]),
                MachineOption(
                    "CONFIG_CORESIGHT_CPU_DEBUG=y", ["aarch64", "arm"]
                ),
                # BPF
                "CONFIG_BPF=y",
                "CONFIG_BPF_JIT=y",
                "CONFIG_HAVE_EBPF_JIT=y",
                "CONFIG_BPF_SYSCALL=y",
            ]
        }

    def _filter_options(self, options) -> List[str]:
        res = []
        for option in options:
            if isinstance(option, str):
                res.append(option)
            else:
                if option.filter(self.board):
                    res.append(option.option)

        return res

    def predefined_configs(
        self,
    ) -> Generator[Tuple[str, List[str]], None, None]:
        """Return all predefined configs that are applicable to a board

        Returns:
           An iterable of tuple of string(name) and list of string [kernel config options]
        """

        for k, v in self._predefined_configs.items():
            yield (k, self._filter_options(v))

    def predefined_config(self, name: str) -> List[str]:
        """Search for predefined config options with given name"""
        for config_name, config in self.predefined_configs():
            if config_name == name:
                return config

        raise PredefinedNotFoundException(
            "Could not find predefined config {}".format(name)
        )

    def predefined_config_fragment(self, name: str) -> str:
        fragment = ""
        try:
            fragment = "\n".join(self.predefined_config(name))
        except PredefinedNotFoundException:
            pass
        return fragment


class KernelData(object):
    """ Provides collection of calculated kernel data """

    def __init__(self, board, kernel_desc):
        self.board = board
        self.kernel_desc = kernel_desc

    @property
    def srcdir(self):
        return self.kernel_desc.kernel_srcdir

    @property
    def arch(self):
        arch = (
            self.board.os.triple.machine.value
            if self.board.os.triple.machine.value != "aarch64"
            else "arm64"
        )
        return arch

    @property
    def shared_data_dir(self):
        """Location for shared kernel sources and cached deploy and build packages"""
        kernel_shared_dir = self.kernel_desc.kernel_source.parent

        return kernel_shared_dir

    @property
    def deploy_package_name(self):
        """Name of the kernel deploy archive"""
        return "kernel-{0}.tar.gz".format(self.kernel_desc.name)

    @property
    def deploy_package_path(self):
        return self.shared_data_dir / self.deploy_package_name

    @property
    def build_cache_name(self):
        """Name of the kernel build cache archive"""
        return "kernel-build-{}.tar.bz".format(self.kernel_desc.name)

    @property
    def build_cache_path(self):
        return self.shared_data_dir / self.build_cache_name



from pathlib import Path

from fabric import task

from ..model.common import Toolchain


@task
def configure(
    c,
    board,
    builddir="",
    srcdir="",
    prefix="",
    flags=None,
    cflags=None,
    cxxflags=None,
    ldflags=None,
    libs=None,
    sysroot=True,
    isa=True,
    uarch=True,
    toolchain="gcc",
    compiler_name="",
    D=[],
):  # pragma: no cover
    """ 
    Configure a cmake project for the build
    
    # Arguments

       -b, --board=STRING: Board for the build. required argument 
       -s, --srcdir=STRING: Directory containing CMakeLists.txt usually top of source directory, required argument
      --builddir=STRING: Directory containing the built project (default: builds/<board_name>)
       --cflags=STRING: Extra flags for C-Compiler
       --flags=STRING: Extra base flags, used for all compilers and linking (for example -flto)
       --libs=STRING: Extra libraries to linke with the binaries
       --ldflags=STRING: Extra flags for linking. Linker is execute through compiler so for actual linker flags use -Wl,--linkerflag
       --compiler-name=STRING: Select a specific compiler 
       --prefix=STRING: Install prefix on the board (default: board.rundir/name of srcdir)
       --toolchain=STRING: Select a different toolchain type choices: llvm, gcc (default: gcc)
       --cxxflags=STRING: Extra flags for C++ compiliation
       --[no-]uarch:  Enable / Disable uarch specific optimizations if uarch is not supported by the compiler 
       --[no-]isa: Enable / Disable isa specific optimizations
       --[no-]sysroot: Enable / Disable linking with default sysroot 
       -D=STRING: Extra definitions for the CMAKE Build e.g. "-D 'ENABLE_OPENCL=ON'"
    
    """

    board = c.board(board)

    toolchain = Toolchain(toolchain) if toolchain else Toolchain.GCC

    cc = board.compiler(toolchain=toolchain, compiler_name=compiler_name)
    cc.configure(
        flags=flags,
        cflags=cflags,
        cxxflags=cxxflags,
        ldflags=ldflags,
        uarch_opt=uarch,
        isa_opt=isa,
        enable_sysroot=sysroot,
        libs=libs,
    )

    builder = board.builder("cmake", builddir=builddir)

    builder.configure(cc, srcdir=srcdir, prefix=prefix, cmake_definitions=D)


@task
def build(c, board, builddir=""):  # pragma: no cover
    """build a cmake project for the board
    
    -b STRING, --board=STRING: Target board
    --builddir=STRING: Builddirectory (default: builds/<board_name>)
    """

    board = c.board(board)
    builder = board.builder("cmake", builddir=builddir)

    builder.build(c)


@task
def install(c, board, builddir=""):  # pragma: no cover
    """install cmake project for deployment in <builddir>/install

    -b STRING, --board=STRING: Target board
    --builddir=STRING: Builddirectory (default: builds/<board_name>)
    """

    board = c.board(board)
    builder = board.builder("cmake", builddir=builddir)

    builder.install(c)


@task
def deploy(c, board, builddir=""):  # pragma: no cover
    """Deploy installed cmake project on board in configured prefix

    -b STRING, --board=STRING: Target board
    --builddir=STRING: Builddirectory (default: builds/<board_name>)
    """

    board = c.board(board)
    builder = board.builder("cmake", builddir=builddir)

    builder.deploy(c)


@task
def clean(c, board, builddir=""):  # pragma: no cover
    """Remove the build directory
    
    -b STRING, --board=STRING: Target board
    --builddir=STRING: Builddirectory (default: builds/<board_name>)
    """

    board = c.board(board)
    builder = board.builder("cmake", builddir=builddir)

    builder.clean(c)


__all__ = ["configure", "build", "clean", "install", "deploy"]



"""
Management of various (usually binary) package types - OS, language, etc.
"""

# TODO: intent is to have various submodules for the various package managers -
# apt/deb, rpm/yum/dnf, arch/pacman, etc etc etc.


from ...patchwork.info import distro_family


def package(c, *packages):
    """
    Installs one or more ``packages`` using the system package manager.

    Specifically, this function calls a package manager like ``apt-get`` or
    ``yum`` once per package given.
    """
    # Try to suppress interactive prompts, assume 'yes' to all questions
    apt = "DEBIAN_FRONTEND=noninteractive apt-get install -y {}"
    # Run from cache vs updating package lists every time; assume 'yes'.
    yum = "yum install -y %s"
    manager = apt if distro_family(c) == "debian" else yum
    for package in packages:
        c.sudo(manager.format(package))


def rubygem(c, gem):
    """
    Install a Ruby gem.
    """
    return c.sudo("gem install -b --no-rdoc --no-ri {}".format(gem))



from datetime import datetime

import tabulate
from invoke import task

from ..utils import connection_to_string


@task
def show_context(c):  # pragma: no cover
    from pprint import pprint

    pprint(dict(c), indent=2)


@task
def list(c, boards=False, compilers=False):  # pragma: no cover
    """List available boards and compilers

      -b/--boards: only list boards
      -c/--compilers: only list compilers
    """

    if not boards and not compilers:
        boards = True
        compilers = True

    if boards:
        board_table = []
        board_header = [
            "Name",
            "Machine",
            "Cores",
            "OS",
            "Connections",
            "Default Compiler",
            "Lock Status",
        ]
        for board in c.boards():
            os = getattr(board.os, "distribution", "unknown")

            connections = []
            for connection in board.connections:
                s = connection_to_string(connection)
                connections.append(s)

            default_compiler_name = ""
            try:
                default_compiler_name = board.compiler().name
            except:
                pass

            lock_status = "unlocked"
            if board.is_locked() or board.has_lock():
                lock_user = board.lock_holder()
                lock_lease_time_end = datetime.now() + board.lock_lease_time()
                lock_status = (
                    f"locked by {lock_user} until {lock_lease_time_end}"
                )

            board_line = [
                board.name,
                board.board,
                len(board.cores),
                os,
                ",".join(connections),
                default_compiler_name,
                lock_status,
            ]

            board_table.append(board_line)

        print("Boards:")
        print(tabulate.tabulate(board_table, headers=board_header))
        print("")

    if compilers:
        print("Compiler:")

        compiler_table = []
        compiler_header = [
            "Name",
            "Toolchain",
            "Version",
            "Machines",
            "Multiarch",
        ]

        for compiler in c.compilers():
            machines = set()
            for triple in compiler.triples:
                machines.add(triple.machine.value)
            compiler_line = [
                compiler.name,
                compiler.toolchain.value,
                compiler.version,
                ", ".join(sorted(machines)),
                "yes" if compiler.multiarch else "no",
            ]
            compiler_table.append(compiler_line)

        print(tabulate.tabulate(compiler_table, headers=compiler_header))
        print("")

    return 0



#!/usr/bin/env python3

import logging
import sys

import coloredlogs
from invoke import Collection, Config, Exit, Program, Result, UnexpectedExit

from . import __version__ as self_version
from . import tasks
from .config import AutomateConfig
from .executor import AutomateExecutor


class AutoTool(Program):
    def core_args(self):
        core_args = super(AutoTool, self).core_args()

        return core_args

    def print_version(self):
        from fabric import __version__ as fabric_version
        from invoke import __version__ as invoke_version
        from paramiko import __version__ as paramiko_version

        print("version: {}".format(self_version))
        print("  fabric: {}".format(fabric_version))
        print("  invoke: {}".format(invoke_version))
        print("  paramiko: {}".format(paramiko_version))

    def execute(self):
        if self.args.debug.value:
            coloredlogs.install(level="DEBUG")
        else:
            if "logging" in self.config and "level" in self.config.logging:
                coloredlogs.install(level=self.config.logging.level)
            else:
                coloredlogs.install(level="WARNING")

        try:
            super(AutoTool, self).execute()
        except UnexpectedExit as e:
            raise e
        except Exit as e:
            raise e
        except Exception as e:
            if not self.args.debug.value:
                result = Result(exited=1)
                logging.fatal("Unexpected Exit: %s", str(e))
                raise UnexpectedExit(result, reason=str(e))
            else:
                raise e


program = AutoTool(
    version=self_version,
    namespace=tasks.collection,
    config_class=AutomateConfig,
    executor_class=AutomateExecutor,
)


program_run = AutoTool(
    version=self_version,
    config_class=AutomateConfig,
    executor_class=AutomateExecutor,
)



from invoke import Collection

from ..database import database_enabled
from . import admin, board, cmake, common, compiler, database, kernel, make

compiler_tasks = Collection().from_module(compiler)
board_tasks = Collection().from_module(board)
admin_tasks = Collection().from_module(admin)
cmake_tasks = Collection().from_module(cmake)
kernel_tasks = Collection().from_module(kernel)
make_tasks = Collection().from_module(make)
database_tasks = Collection().from_module(database)

collection = Collection().from_module(common)
collection.add_collection(compiler_tasks)
collection.add_collection(board_tasks)
collection.add_collection(admin_tasks)
collection.add_collection(cmake_tasks)
collection.add_collection(kernel_tasks)
collection.add_collection(make_tasks)

if database_enabled():
    collection.add_collection(database_tasks)



from pathlib import Path
from string import Template

from ..model.board import UBootModel

FIT_TEMPLATE = Template(
    r"""
/*
 * Simple U-Boot uImage source file containing a single kernel and FDT blob
 */

/dts-v1/;

/ {
        description = "Simple image with single Linux kernel and FDT blob";
        #address-cells = <1>;

        images {
                kernel {
                        description = "Linux kernel";
                        data = /incbin/("${kernel_image}");
                        type = "kernel";
                        arch = "${arch}";
                        os = "linux";
                        compression = "${image_compression}";
                        load = <${loadaddr}>;
                        entry = <${loadaddr}>;
                        hash-1 {
                                algo = "crc32";
                        };
                        hash-2 {
                                algo = "sha1";
                        };
                };
                fdt-1 {
                        description = "Flattened Device Tree blob";
                        data = /incbin/("${dtb_image}");
                        type = "flat_dt";
                        arch = "${arch}";
                        compression = "${dtb_compression}";
                        hash-1 {
                                algo = "crc32";
                        };
                        hash-2 {
                                algo = "sha1";
                        };
                };
        };

        configurations {
                default = "conf-1";
                conf-1 {
                        description = "Boot Linux kernel with FDT blob";
                        kernel = "kernel";
                        fdt = "fdt-1";
                };
        };
};
"""
)


def build_ubimage(
    c,
    uboot_desc: UBootModel,
    arch: str,
    build_path: Path,
    boot_path: Path,
    kernel_image: Path,
) -> None:
    """ Build a U-Boot image file  for kernel deployment

    # Arguments
    c: AutomateContext to run commands on
    uboot_desc: UBootModel describing image configuration
    build_path: Path of kernel Build directory
    boot_path: path of built ubimage (usually install/boot inside kernel build directory)
    kernel_image: path to kernel_image file that should be packaged inside builddir
    """
    loadaddr = uboot_desc.loadaddr
    image_name = uboot_desc.image_name
    dtb_image = uboot_desc.dtb_image
    image_compression = uboot_desc.image_compression
    dtb_compression = uboot_desc.dtb_compression

    result = FIT_TEMPLATE.safe_substitute(
        {
            "loadaddr": loadaddr,
            "image_name": str(image_name),
            "dtb_image": str(dtb_image),
            "arch": arch,
            "kernel_image": str(kernel_image),
            "image_compression": str(image_compression),
            "dtb_compression": str(dtb_compression),
        }
    )

    fit_path = build_path / "fit_image.its"
    image_path = boot_path / image_name

    with fit_path.open("w") as f:
        f.write(result)

    with c.cd(str(build_path)):
        c.run("mkimage -f {0} {1}".format(str(fit_path), str(image_path)))



from ._version import __version__, __version_info__
from .board import Board
from .compiler import Compiler, CrossCompiler
from .config import AutomateConfig
from .context import AutomateContext

__all__ = [
    "__version_info__",
    "__version__",
    "AutomateConfig",
    "AutomateContext",
    "Board",
    "Compiler",
    "CrossCompiler",
]



import logging
import os
import tarfile
from pathlib import Path
from typing import List, Union

from ..model.board import SSHConnectionModel, UARTConnectionModel


def connection_to_string(
    connection: Union[SSHConnectionModel, UARTConnectionModel]
) -> str:
    table = {SSHConnectionModel: "ssh", UARTConnectionModel: "uart"}

    name = "UNKNOWN"

    t = type(connection)

    if t in table:
        name = table[t]

    return name


def fix_symlinks(base_path: Path) -> None:
    """ This function tries to fix symlinks in cloned buildroots by:

1. Searching for all symlinks with absolute target path in base_path
2. Prepending base_path to the target path
3. Making the targets of the links relative to the symlink location
4. Deleting the symlink and replacing it with one to the relative location
 """

    logging.info("Changing absolute symlinks to relative symlinks")
    links: List[str] = []
    for root, _, files in os.walk(base_path):
        for filename in files:
            path = os.path.join(root, filename)
            if os.path.islink(path):
                links.append(path)
            else:
                # If it's not a symlink we're not interested.
                continue

    for link in links:
        target = os.readlink(link)
        if os.path.isabs(target):
            new_target = str(base_path) + str(target)
            new_target_rel = os.path.relpath(
                new_target, os.path.dirname(os.path.abspath(link))
            )

            logging.debug(
                "Relinking\n  link: {}\n  target: {}\n  new_target: {}\n  new_target_rel: {}\n  base_path: {}".format(
                    link, target, new_target, new_target_rel, base_path
                )
            )

            os.unlink(link)
            os.symlink(new_target_rel, link)


def untar(
    tar_path: Union[str, Path], extract_path: Union[str, Path] = "."
) -> None:
    tar_path = Path(tar_path)
    extract_path = Path(extract_path)

    if not extract_path.exists():
        extract_path.mkdir(parents=True)

    with tarfile.open(name=tar_path, mode="r|*") as tar_file:
        tar_file.extractall(extract_path)



import concurrent
import logging
import os.path
import re
import shlex
import shutil
import subprocess
import tempfile
from collections import namedtuple
from enum import Enum
from pathlib import Path

import automate._vendor.patchwork.files
from fabric import Connection, task
from paramiko.ssh_exception import AuthenticationException
from prompt_toolkit import prompt
from ruamel.yaml import YAML  # type: ignore

from ..loader import ModelLoader
from ..model import (
    BoardModel,
    CoreModel,
    GatewayModel,
    OSModel,
    SSHConnectionModel,
    TripleModel,
)
from ..utils import cpuinfo, fix_symlinks
from ..utils.network import connect, find_local_port


@task
def add_users(c):  # pragma: no cover
    """Add users ssh keys to all boards and gateways
    """
    loader = ModelLoader(c.config)
    users = loader.load_users()

    def copy_keys(sftp, users, homedir):
        patchwork.files.directory(con, "~/.ssh", mode="700")

        authorized_keys = homedir / ".ssh" / "authorized_keys"
        with sftp.open(str(authorized_keys), "w") as authorized_keys_file:
            for user_id, user in users.users.items():
                authorized_keys_file.write("# {}\n".format(user_id))
                for ssh_key in user.public_keys:
                    ssh_key = ssh_key.strip()
                    authorized_keys_file.write("{}\n".format(ssh_key))

    FrozenGateway = namedtuple("FrozenGateway", ["host", "username", "port"])

    gateways = set()

    for board in c.boards():
        if board.gateway is not None:
            gateways.add(FrozenGateway(**board.gateway.dict()))

        try:
            with board.connect() as con:

                sftp = con.sftp()
                homedir = board.homedir()

                copy_keys(sftp, users, homedir)
        except Exception as e:
            logging.error(
                f"Could not connect to board {board.name} please try again later"
            )

    for gw in gateways:
        identity = os.path.expanduser(c.config.automate.identity)
        try:

            with Connection(
                host=gw.host,
                user=gw.username,
                port=gw.port,
                connect_kwargs={"key_filename": identity},
            ) as con:

                result = con.run("echo $HOME", hide=True)
                gw_homedir = Path(result.stdout.strip())
                sftp = con.sftp()
                copy_keys(sftp, users, gw_homedir)

        except Exception as e:
            logging.error(
                f"Could not connect to gw {gw.host} please try again later"
            )


@task
def safe_rootfs(c, board):  # pragma: no cover
    """Safe rootfs image of board
       
        -b/--board: target board name
    """  # noqa
    bh = c.board(board)

    port = find_local_port()

    image_name = Path(bh.os.rootfs)
    image_name.parent.mkdir(parents=True, exist_ok=True)

    print("Cloning rootfs to {}\n".format(image_name))

    with bh.lock_ctx():

        logging.info("Connecting to target using port {}".format(port))

        con = bh.connect()

        result = con.run(
            "echo 1 | sudo tee /proc/sys/kernel/sysrq", hide="stdout", pty=True
        )
        if result.return_code != 0:
            raise Exception("Could not enable sysrq triggers")

        mount_result = con.run("mount", hide="stdout")
        if mount_result.return_code != 0:
            raise Exception("Could not get mountpoints")
        mount_output = mount_result.stdout
        mount_table_pattern = re.compile(r"(.*) on (.*) type (.*) \((.*)\)")
        rootdevice = ""
        for line in mount_output.splitlines():
            line = line.strip()
            match = re.match(mount_table_pattern, line)
            if match:
                device = match.group(1).strip()
                mountpoint = match.group(2).strip()
                # fstype = match.group(3).strip()
                # args = match.group(4).strip()

                if mountpoint == "/":
                    rootdevice = device
                    break

        if not rootdevice:
            raise Exception("Could not find root device for {}".format(board))

        logging.info("Using device: {}".format(rootdevice))

        with concurrent.futures.ThreadPoolExecutor() as thread_executor:
            try:
                with open(image_name.with_suffix(".tmp"), "wb") as image_file:
                    logging.info("Starting listener")

                    def reader_func():

                        reader_cmd = "nc -l {}".format(port)
                        reader = subprocess.Popen(
                            shlex.split(reader_cmd), stdout=subprocess.PIPE
                        )

                        shutil.copyfileobj(reader.stdout, image_file)
                        return reader.wait()

                    reader_result = thread_executor.submit(reader_func)

                    result = con.run(
                        "echo u | sudo tee /proc/sysrq-trigger",
                        hide="stdout",
                        pty=True,
                    )
                    if result.return_code != 0:
                        raise Exception(
                            "Could not remount file systems read only"
                        )

                    with con.forward_remote(port):
                        logging.info("Starting writer")
                        res = con.run(
                            "sudo dd if={} |  nc -N localhost {}".format(
                                rootdevice, port
                            ),
                            pty=True,
                        )

                        logging.info("waiting for reader")

                        result = reader_result.result()
                        if result != 0:
                            raise Exception("Could not write image!")

                        if res.return_code != 0:
                            raise Exception("Error during image writing")

            except BaseException as e:
                if image_name.with_suffix(".tmp").exists():
                    c.run("rm {}".format(image_name.with_suffix(".tmp")))
                logging.error(
                    "Exception during image writing: {}".format(str(e))
                )
            finally:
                logging.info("Rebooting target {}".format(board))
                bh.reboot(wait=False)

        logging.info("Backup existing image")
        if image_name.exists():
            c.run(
                "mv {0} {1}".format(image_name, image_name.with_suffix(".bak"))
            )

        logging.info("Fsck image")
        c.run("fsck -p {}".format(image_name.with_suffix(".tmp")))

        logging.info("Moving image to result")
        c.run("mv {0} {1}".format(image_name.with_suffix(".tmp"), image_name))
        print("Finished image saving.")

        return 0


@task
def build_sysroot(c, board):  # pragma: no cover
    """Build compiler sysroot for board
       
        -b/--board: target board id
    """  # noqa

    bh = c.board(board)

    image_path = Path(bh.os.rootfs)
    if not image_path.exists():
        safe_rootfs(c, board)

    try:
        tmp_path = Path(tempfile.mkdtemp())
        logging.debug("Using mountpoint {}".format(tmp_path))

        c.run("sudo mount -l {} {}".format(str(image_path), str(tmp_path)))

        bh.os.sysroot.mkdir(exist_ok=True, parents=True)

        try:
            rsync_result = c.run(
                r'rsync -ar --delete --delete-excluded --exclude="/tmp" --exclude="/home" {}/ {}/'.format(
                    str(tmp_path), str(bh.os.sysroot)
                ),
                hide="stdout",
                warn=True,
            )

            if rsync_result.return_code != 0:
                logging.warning("Rootfs might be incomplete")

            fix_symlinks(bh.os.sysroot)

        except BaseException as e:
            print(e)
            raise e
        finally:
            c.run("sudo umount {}".format(tmp_path))

    except BaseException as e:
        print(e)
        raise e
    finally:
        c.run("sudo rmdir {}".format(tmp_path))


board_yaml_template = r"""
name: 
hostname: 
board: 
description: 
rundir:

# Links to documentation of these boards
doc:

# Gateway used for this board
gateway:
  host:
  username:

# Available UART and SSH Connections
connections:
  - 
    host:
    username:

# Available Cores
cores:
  -
    id:
    description: 
    isa:
    uarch: 
    vendor: 

# Description of the Board OS and the available kernels
os:
  triple:
    os: 
    machine: 
    environment:
    distribution:
    release: 
    description: 
    sysroot:
    rootfs: 
    multiarch: 
    kernels: 
      - 
        id: 
        description: 
        version: 
        commandline: 
        kernel_srcdir: 
        kernel_config: 
        kernel_source: 
        default:
"""  # noqa


@task
def add_board(
    c, user="", host="", port=22, gw_host="", gw_user="", gw_port=22
):  # pragma: no cover
    """Add a new board to test rack

      # Arguments
      --user: username on board
      --host: hostname or ip address of board
      --port: port of ssh deamon on board (optional default: 22)
      --gw-host: hostname of gateway (optional if omitted not gateway is configured)
      --gw-user: username on gateway (optional if omitted use --user)
      --gw-port: port of ssh on gateway (optional default: 22)
    """  # noqa

    if host == "":
        raise Exception("No hostname given: --host missing")

    if user == "":
        raise Exception("No username given --user missing")

    gateway_connection = None
    if gw_host != "":
        if gw_user == "":
            gw_user = user

        gateway_connection = connect(
            gw_host,
            gw_user,
            gw_port,
            passwd_allowed=True,
            keyring_allowed=False,
            gateway=gateway_connection,
        )

    try:
        con = connect(
            host,
            user,
            port,
            passwd_allowed=False,
            keyring_allowed=False,
            gateway=gateway_connection,
        )
        con.open()
    except AuthenticationException:
        print("Could not Authenticate with public key")

        con = connect(
            host,
            user,
            port,
            passwd_allowed=True,
            keyring_allowed=False,
            gateway=gateway_connection,
        )
        con.open()

        identity = os.path.expanduser(c.config.automate.identity)
        keyfile = prompt(
            "SSH Public key file: ",
            default=str(identity) + ".pub",
            is_password=False,
        )

        with open(keyfile) as f:
            key = f.read()

            patchwork.files.directory(con, "~/.ssh", mode="700")
            patchwork.files.append(con, "~/.ssh/authorized_keys", key)

            con.close()

            con = connect(
                host,
                user,
                port,
                passwd_allowed=True,
                keyring_allowed=False,
                gateway=gateway_connection,
            )

            con.open()

    assert con.is_connected

    result = con.run("hostname", hide="stdout", warn=True)
    hostname = ""
    if result.return_code == 0:
        hostname = result.stdout.strip()

    board_name = prompt("board_name: ", default=hostname)
    model_file = (
        Path(c.config.automate.metadata)
        / "boards"
        / board_name
        / "description.yml"
    )
    if model_file.exists():
        logging.error("board with id {0} already exists".format(board_name))
        return -1

    board_model = board_name
    board_model = prompt("board model: ", default=board_model)

    board_description = prompt("board description: ", default="")

    rundir = "/home/{}/run".format(user)
    rundir = prompt("board rundir: ", default=rundir)

    cpus = cpuinfo.cpuinfo(con)
    cpu_models = []
    # TODO: this is quite repetitive if there are many cores ;)
    for cpu in cpus:
        print("cpu: ", cpu.num)

        description = prompt(
            "  description: ", validator=None, default=cpu.description
        )
        uarch = prompt("  microarchitecture: ", default=cpu.uarch)

        vendor = prompt("  vendor: ", default=cpu.vendor)

        isa = prompt("  isa: ", default=cpu.isa)

        # TODO: Prompt for isa extensions

        cpu_model = CoreModel(
            os_id=cpu.num,
            isa=isa,
            uarch=uarch,
            vendor=vendor,
            extensions=cpu.extensions,
            description=description,
        )
        cpu_models.append(cpu_model)

    print("OS Configuration")

    t_machine = "arm64"
    result = con.run("uname --machine", hide="stdout", warn=True)
    if result.return_code == 0:
        t_machine = result.stdout.strip()

    machine_sentinels = {
        "armv7l": "arm",
        "arm": "arm",
        "aarch64": "aarch64",
        "arm64": "aarch64",
    }

    if t_machine in machine_sentinels:
        t_machine = machine_sentinels[t_machine]

    t_abi = "gnueabihf"
    if t_machine == "aarch64":
        t_abi = "gnu"

    triple = prompt(
        "  triple (machine-vendor-os-vendor): ",
        default="{}-unknown-linux-{}".format(t_machine, t_abi),
    )
    t_machine, t_vendor, t_os, t_abi = triple.split("-")
    triple = TripleModel(
        machine=t_machine, vendor=t_vendor, os=t_os, environment=t_abi
    )

    result = con.run("cat /etc/os-release", hide="stdout", warn=True)
    distribution = ""
    release = ""
    if result.return_code == 0:
        for line in result.stdout.split("\n"):
            line = line.strip()
            if len(line.split("=")) != 2:
                continue
            k, v = line.split("=")
            v = v.strip('"')

            if k == "ID":
                distribution = v
            elif k == "VERSION_ID":
                version = v

    distribution = prompt("  distribution: ", default=distribution)
    release = prompt("  release: ", default=release)
    description = ""
    sysroot = prompt(
        "  sysroot: ", default="${boardroot}/${board_name}/sysroot"
    )
    rootfs = prompt(
        "  rootfs: ", default="${boardroot}/${board_name}/${board_name}.img"
    )

    multiarch = False
    if distribution in ["ubuntu", "debian", "raspbian"]:
        multiarch = True

    # TODO: extract kernel info
    os_model = OSModel(
        triple=triple,
        distribution=distribution,
        release=release,
        description=description,
        sysroot=sysroot,
        rootfs=rootfs,
        multiarch=multiarch,
    )

    gateway = None
    if gw_host:
        gateway = GatewayModel(host=gw_host, username=gw_user, port=gw_port)

    board_model = BoardModel(
        name=board_name,
        description=board_description,
        board=board_model,
        gateway=gateway,
        rundir=rundir,
        connections=[SSHConnectionModel(username=user, host=host, port=port)],
        cores=cpu_models,
        os=os_model,
    )

    model_file.parent.mkdir(parents=True, exist_ok=True)

    yaml = YAML(typ="rt")
    yaml.default_flow_style = False
    yaml.indent(mapping=2, sequence=4, offset=2)

    yaml.load(board_yaml_template)
    with model_file.open("w") as mf:
        d = board_model.dict()

        def _recurse(d):
            if isinstance(d, dict):
                for k, v in d.items():
                    d[k] = _recurse(v)
                return d
            elif isinstance(d, list):
                return [_recurse(item) for item in d]
            elif isinstance(d, Enum):
                return d.value
            elif isinstance(d, Path):
                return str(d)
            else:
                return d

        d = _recurse(d)
        yaml.dump(d, mf)


@task
def deploy_runtimes(c, board):
    """ Deploy compiler runtimes to boards

    Currently deploys runtime of default gcc compiler

    #Arguments
    -b/--board: name of board to deploy to
    """

    boards = []
    if board == "all":
        boards = list(c.boards())
    else:
        boards.append(c.board(board))

    while boards:
        board = boards[0]
        boards.pop()
        if board.is_locked():
            boards.append(board)
        else:
            with board.lock_ctx():
                with board.connect() as con:
                    compiler = board.compiler()
                    if not compiler.runtime:
                        logging.warn(
                            "Could not find compiler runtime for %s",
                            compiler.name,
                        )
                        continue
                    con.run("sudo mkdir -p /opt/runtimes")
                    con.put(compiler.runtime, "/tmp")
                    with con.cd("/opt/runtimes"):
                        con.run(
                            f"sudo mv /tmp/{Path(compiler.runtime).name} /opt/runtimes"
                        )
                        result = con.run(
                            f"sudo tar xvJf {Path(compiler.runtime).name}"
                        )
                        con.run(f"sudo rm -rf {Path(compiler.runtime).name}")

                        lib_dirs = set()
                        for line in result.stdout.splitlines():
                            file_path = Path(line)
                            print(line)
                            if file_path.suffix == ".so":
                                lib_dirs.add(str(file_path.parent))

                        if len(lib_dirs) < 1:
                            continue

                        runtime_name = (
                            "00_"
                            + str(Path(list(lib_dirs)[0]).parts[0])
                            + ".conf"
                        )

                        sftp = con.sftp()

                        sftp.chdir("/tmp")
                        remote_file = sftp.file(runtime_name, mode="w")
                        remote_file.write(
                            f"#compiler runtime for {compiler.name}\n"
                        )

                        for path in lib_dirs:
                            full_path = Path("/opt/runtimes") / path
                            remote_file.write(f"{full_path}\n")

                        con.run(
                            f"sudo cp /tmp/{runtime_name} /etc/ld.so.conf.d/"
                        )
                        con.run("sudo ldconfig")



import logging
import random
import socket
import time
from io import StringIO
from pathlib import Path
from typing import Iterable, Optional

import fabric
import keyring
from paramiko.ssh_exception import AuthenticationException, ChannelException
from prompt_toolkit import prompt

from ..locks import KeepLockThread


class GatewayManagingConnection(fabric.Connection):
    def __init__(
        self,
        host,
        user=None,
        port=None,
        config=None,
        gateway=None,
        forward_agent=None,
        connect_timeout=None,
        connect_kwargs=None,
        inline_ssh_env=None,
    ):
        super().__init__(
            host,
            user,
            port,
            config,
            gateway,
            forward_agent,
            connect_timeout,
            connect_kwargs,
            inline_ssh_env,
        )
        self.gateway = gateway

    def __enter__(self, *args, **kwargs):
        return super().__enter__(*args, **kwargs)

    def __exit__(self, *args, **kwargs):
        self.close()
        return super().__exit__(*args, **kwargs)

    def close(self):
        if self.gateway is not None:
            self.gateway.close()
            self.gateway = None

        return super().close()

    def __del__(self):
        self.close()


def connect(
    host: str,
    user: str,
    port: int = 22,
    identity: Optional[Path] = None,
    passwd_allowed: bool = False,
    passwd_retries: int = 3,
    keyring_allowed: bool = True,
    gateway: Optional[fabric.Connection] = None,
    timeout: int = 30,
) -> fabric.Connection:
    """ Get a fabric connection to a remote host 

        # Arguments
        host: hostname or ip address
        username: on remote host
        port: ssh port for connection
        identity: Path to ssh private key
        passwd_allowed: if True use password if ssh public key authentication fails
        passwd_retries: number of retries for password authentication
        keyring_allowed: if True store passwords in system keyring
        gateway: fabric.Connection to use as gateway
        timeout: timeout for connections in seconds

        # Returns
        a fabric.Connection to the host
    """  # noqa

    try:
        kwargs = {"key_filename": str(identity.absolute())} if identity else {}
        connection = GatewayManagingConnection(
            host,
            user=user,
            port=port,
            connect_timeout=timeout,
            gateway=gateway,
            connect_kwargs=kwargs,
        )
        connection.open()
    except AuthenticationException as e:
        if passwd_allowed:

            keyring_service = f"automatessh:{host}:{port}"

            for retry in range(passwd_retries):
                password = None
                if retry == 0 and keyring_allowed:
                    password = keyring.get_password(keyring_service, user)
                if password is None:
                    password = prompt(
                        "Password for {}@{}: ".format(user, host),
                        is_password=True,
                    )
                try:
                    connection = GatewayManagingConnection(
                        user=user,
                        host=host,
                        port=port,
                        gateway=gateway,
                        connect_timeout=timeout,
                        connect_kwargs={"password": password},
                    )
                    connection.open()
                except AuthenticationException:
                    continue

                if keyring_allowed:
                    keyring.set_password(keyring_service, user, password)
                break
        else:
            raise e

    return connection


def find_local_port(start=1024, end=65536) -> int:
    """ Returns a locally bindable port number 

    # Returns
    port number [int]
    """  # noqa

    while True:
        port = random.randint(start, end)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            logging.debug("selected local port: %d", port)
            sock.bind(("0.0.0.0", port))
            sock.close()
            return port
        except Exception as e:
            logging.debug("Port {} is not bindable".format(port))


def find_remote_port(con) -> int:
    """ Returns a port number bindable on the remote end

    # Returns
    port number [int]
    """  # noqa

    while True:
        port = random.randint(1024, 65536)
        result = con.run(f"nc -zv localhost {port}", hide="both", warn=True)
        if result.exited != 0:
            logging.debug("selected remote port: %d", port)
            return port
        logging.info("Command nc exited with %s", str(result.exited))


RSYNC_SPEC = """
port={port}
use chroot=false
log file=/tmp/rsync-ad-hoc.{id}.log
pid file=/tmp/rsync-ad-hoc.{id}.pid
[files]
max verbosity=4
path=/
read only=false
munge symlinks=false
"""


def rsync(
    con: fabric.Connection,
    source: str,
    target: str,
    exclude: Iterable[str] = (),
    delete: bool = False,
    verbose: bool = False,
    rsync_timeout: int = 20,
    retries: int = 5,
    rsync_opts: str = "",
) -> None:
    """ RSync files or folders to board 

    1. Starts a remote rsync forwards
    2. Forwards rsync server ports over gateway
    3. runs rsync -pthrz <source> <target>
    4. stops remote rsync daemon

    rsync server is run as the connections default user, so can not modify files and folders for which this user does not have access rights 

    # Arguments
    con: fabric.Connection to board
    source: local path should end in "/" if the complete folder is synced
    target: remote_path
    exclude: iterable of exclude patterns
    rsync_timeout: --timeout argument for rsync
    retries: number of retries if rsync fails
    verbose: if True print transfered files to stdout
    rsync_opts: string of additional rsync options
    """  # noqa
    retry = True
    while retry and retries > 0:
        retry = False
        retries -= 1
        rsync_id = random.randint(0, 2 ** 31)
        local_port = find_local_port()
        remote_port = find_remote_port(con)
        logging.info("Starting rsync daemon on port: %d", remote_port)
        try:
            with con.forward_local(local_port, remote_port):
                try:
                    con.put(
                        StringIO(
                            RSYNC_SPEC.format(port=remote_port, id=rsync_id)
                        ),
                        f"/tmp/rsync-ad-hoc.{rsync_id}.conf",
                    )

                    con.run(
                        f"rsync --daemon --config /tmp/rsync-ad-hoc.{rsync_id}.conf"
                    )
                    con.run(f"mkdir -p {target}")

                    delete_flag = "--delete" if delete else ""

                    exclude_opts = " ".join(
                        ["--exclude %s" % e for e in exclude]
                    )
                    if verbose:
                        rsync_opts = "-v " + rsync_opts

                    remote_path = (
                        f"rsync://localhost:{local_port}/files/{target}"
                    )
                    rsync_cmd = f"rsync --timeout {rsync_timeout} {delete_flag} {exclude_opts} -pthrz {rsync_opts} {source} {remote_path}"
                    logging.info("Running {}".format(rsync_cmd))
                    con.local(rsync_cmd)
                except Exception as e:
                    print(e)
                    raise (e)
                finally:
                    result = con.run(
                        f"cat /tmp/rsync-ad-hoc.{rsync_id}.pid", hide="out"
                    )
                    rsync_pid = result.stdout
                    logging.info(
                        f"Killing remote rsync deamon with pid: {rsync_pid}"
                    )
                    con.run(f"kill  {rsync_pid}", hide="out")

                    con.run("rm -f /tmp/rsync-ad-hoc.{rsync_id}.*")
        except Exception as e:
            retry = True
            if retries == 0:
                raise e

            logging.critical("Channel exception during rsync retrying %s")
            logging.debug("Exception: %s", str(e))
            time.sleep(0.5)



import io
import logging
import os
import string
from datetime import datetime
from glob import glob
from pathlib import Path
from typing import Any, Dict, List, Optional

from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap

from .config import AutomateConfig
from .model import (
    BoardModel,
    BoardModelFS,
    CompilerModel,
    DataModelBase,
    LoadedModelBase,
    MetadataModel,
    UsersModel,
    VersionString,
)


class ModelLoader(object):
    def __init__(self, config: AutomateConfig, database=None) -> None:
        """Initialize model loader

        # Parameters
        config: AutomateConfig object to use for search of metadata locations
        database: optional database connnection
        """

        self.config = config
        self.database = database

        self.logger = logging.getLogger(__name__)
        self.logger.debug(
            "Metadata Loader for {}".format(self.config.automate.metadata)
        )
        self.yaml = YAML()

        self.model = None

    def _load_metadata_list(
        self, pattern: str, recursive: bool = True
    ) -> List[CommentedMap]:
        """Parse yaml files for metadata information

        # Arguments
        pattern: glob_pattern to search for metadata files
        recursive: wether search allows ** in glob patterns to recurse into subdirectories
        """
        res = []
        glob_pattern = os.path.join(
            os.path.expanduser(self.config.automate.metadata), pattern
        )
        self.logger.debug("Load glob pattern: {}".format(str(glob_pattern)))
        files = glob(glob_pattern, recursive=recursive)

        for file_name in files:
            self.logger.debug("Loading metadata from {}".format(file_name))
            with io.open(file_name, encoding="utf-8") as f:
                mtime = datetime.utcfromtimestamp(os.path.getmtime(file_name))

                yaml_dict = self.yaml.load(f)
                yaml_dict["model_file"] = file_name
                yaml_dict["model_file_mtime"] = mtime
                res.append(yaml_dict)

        return res

    def _merge_metadata(self, *args):
        """ Merge multiple lists of Models """
        merged_list = []
        merged_names = set()
        for arg in args:
            for item in arg:
                if item.name in merged_names:
                    continue
                merged_list.append(item)
                merged_names.add(item.name)

        merged_list.sort(key=lambda x: x.name)
        return merged_list

    def _apply_templates(
        self, data_model: DataModelBase, env: Dict[str, str]
    ) -> None:
        """ Expand template parameters from env in strings from data_model """

        env = dict(env)
        env.update(data_model._get_env_dict())

        def do_apply_template(template, env):
            try:
                formatted = template.safe_substitute(env)
                return formatted
            except ValueError as e:  # pragma: no cover
                self.logger.error(str(e))
                self.logger.error(
                    "During formatting of field {} from {}".format(
                        field_name, field
                    )
                )
                self.logger.error(str(env))
                raise e

        for field_name in data_model.__fields__:
            field = getattr(data_model, field_name)

            formatted = ""
            if isinstance(field, str) and not isinstance(field, VersionString):
                template = string.Template(field)
                formatted = do_apply_template(template, env)
                setattr(data_model, field_name, formatted)
            elif isinstance(field, Path):

                template = string.Template(str(field))
                formatted = do_apply_template(template, env)

                formatted_path = Path(formatted)

                setattr(data_model, field_name, formatted_path)

            elif isinstance(field, DataModelBase):
                self._apply_templates(field, env)
            elif isinstance(field, list):
                for item in field:
                    if isinstance(item, DataModelBase):
                        self._apply_templates(item, env)

        return None

    def load(self, expand_templates=True) -> MetadataModel:
        """Load Metadata model

        # Arguments:
            expand_templates: boolean if true ${var} templates Paths are replaced by their respective configuration value

        # Templates
        Expand templates currently expands the following templates
        ${metadata}: metadata location from config
        ${toolroot}: toolroot from config  used to give relative paths for compilers
        ${boardroot}: boardroot from config used to store kernel_sources, rootfs_images, board sysroots, cached_builds
        """
        compiler_map = self._load_metadata_list("compilers/**/description.yml")
        compilers = [CompilerModel(**c) for c in compiler_map]

        board_dicts = self._load_metadata_list("boards/**/description.yml")
        boards = []
        for board_dict in board_dicts:
            try:
                board = BoardModelFS(**board_dict)
                boards.append(board)
            except Exception as e:
                self.logger.error(
                    "Could not validate board description from: %s",
                    str(board_dict["model_file"]),
                )
                self.logger.error(str(e))

        if self.database:
            self.logger.info("getting boards from database")
            database_boards = self.database.get_all_boards()
            self.logger.info(
                "Boards from database %s",
                " ".join((b.name for b in database_boards)),
            )
            boards = self._merge_metadata(boards, database_boards)

        data_model = MetadataModel(compilers=compilers, boards=boards)

        if expand_templates:
            self.logger.info("Expanding templates in metadata files")
            template_env = {
                "metadata": os.path.expanduser(
                    str(self.config.automate.metadata)
                ),
                "toolroot": os.path.expanduser(
                    str(self.config.automate.toolroot)
                ),
                "boardroot": os.path.expanduser(
                    str(self.config.automate.boardroot)
                ),
            }

            self._apply_templates(data_model, template_env)

        return data_model

    def load_users(self) -> UsersModel:
        """ Load user data """
        metadata_path = Path(
            os.path.expanduser(str(self.config.automate.metadata))
        )
        users_file = metadata_path / "users.yml"

        with io.open(users_file, encoding="utf-8") as f:
            mtime = datetime.utcfromtimestamp(os.path.getmtime(users_file))
            yaml_dict = self.yaml.load(f)

            users_model = UsersModel(
                users=dict(yaml_dict),
                model_file=str(users_file),
                model_file_mtime=mtime,
            )

        return users_model



import re
from typing import Any, Dict, List

from fabric import Connection

from ..model import CoreModel
from .cpuinfo_arm import implementers as arm_implementers
from .cpuinfo_arm import uarch_to_isa

__all__ = ["cpuinfo"]


def _cpuinfo(text: str) -> List[CoreModel]:
    """Parse a /proc/cpuinfo file to a list of core models

    # Arguments
    text: contents of cpuinfo file

    # Returns
    List of Core Models for the cpus in cpuinfo
    """

    lines = text.split("\n")

    current_dict: Dict[str, Any] = {}
    cpus: List[CoreModel] = []
    for line in lines:
        m = re.match(r"processor\s+: (\d+)", line)
        if m:
            if current_dict:
                if "isa" not in current_dict or not current_dict["isa"]:
                    if current_dict["uarch"] in uarch_to_isa:
                        current_dict["isa"] = uarch_to_isa[
                            current_dict["uarch"]
                        ]
                current_dict["description"] = current_dict[
                    "description"
                ] + " (microarchitecture: {})".format(current_dict["uarch"])
                cpus.append(CoreModel(**current_dict))
            current_dict = {"description": ""}
            current_dict["num"] = int(m.group(1))
            current_dict["isa"] = ""

        m = re.match(r"model name\s+: (.*)", line)
        if m:
            current_dict["description"] = str(m.group(1)).strip()

        m = re.match(r"CPU implementer\s*: (\S+)", line)
        if m:
            implementer_key = int(m.group(1), 16)

            vendor = ""
            if implementer_key in arm_implementers:
                vendor = arm_implementers[implementer_key][0]

            current_dict["vendor"] = vendor

        m = re.match(r"CPU part\s*: (\S+)", line)
        if m:
            part_key = int(m.group(1), 16)

            uarch = ""
            if implementer_key in arm_implementers:
                if part_key in arm_implementers[implementer_key][1]:
                    uarch = arm_implementers[implementer_key][1][part_key]

            current_dict["uarch"] = uarch

        m = re.match(r"Features\s*: (.*)", line)
        if m:
            features = m.group(1).split()

            current_dict["extensions"] = list(sorted(features))

    if current_dict:

        if "isa" not in current_dict or not current_dict["isa"]:
            if current_dict["uarch"] in uarch_to_isa:
                current_dict["isa"] = uarch_to_isa[current_dict["uarch"]]
        current_dict["description"] = current_dict[
            "description"
        ] + " (microarchitecture: {})".format(current_dict["uarch"])
        cpus.append(CoreModel(**current_dict))

    return cpus


def cpuinfo(con: Connection) -> List[CoreModel]:
    """Parse remote CPU info over ssh connection

    # Arguments
    con: fabric.Connection for the

    # Returns
    A list of CoreModel with the parsed cpuinfos
    """
    result = con.run("cat /proc/cpuinfo", hide="stdout", warn=True)
    if result.return_code != 0:
        return []

    return _cpuinfo(result.stdout)



from datetime import datetime
from pathlib import Path
from typing import Dict

from pydantic import BaseModel


class DataModelBase(BaseModel):
    class Config:
        anystr_strip_whitespace = True
        validate_all = True
        extra = "forbid"
        allow_mutation = True
        allow_population_by_field_name = True

    def _get_env_dict(self) -> Dict[str, str]:
        return {}


class LoadedModelBase(DataModelBase):
    model_file: Path
    model_file_mtime: datetime

    def _get_env_dict(self) -> Dict[str, str]:
        return {
            "model_file": str(self.model_file),
            "model_dir": str(self.model_file.parent),
        }


class DBModelBase(DataModelBase):
    id: int  # Database ID

    class Config(DataModelBase.Config):
        orm_mode = True



import os.path
from pathlib import Path
from typing import Dict, Optional

from .. import compiler
from ..utils import untar
from ..utils.kernel import KernelConfigBuilder, KernelData
from ..utils.network import rsync
from ..utils.uboot import build_ubimage
from .builder import BaseBuilder, BuilderState


class KernelBuilder(BaseBuilder):
    def _kernel_desc(self):
        board = self.board

        kernel_desc = None
        for kernel in board.os.kernels:
            if kernel.name == self._kernel_name():
                kernel_desc = kernel
                break

        if kernel_desc is None:
            raise Exception(
                "Could not find config with id: {} for board {}".format(
                    self._kernel_name(), board.name
                )
            )

        return kernel_desc

    def _kernel_data(self) -> KernelData:
        """Computed kernel data"""

        return KernelData(self.board, self._kernel_desc())

    def _kernel_name(self) -> str:
        """Name for current kernel config"""

        assert self.state.kernel is not None
        return str(self.state.kernel["kernel_name"])

    def _arch(self) -> str:
        """Arch argument"""

        assert self.state.kernel is not None
        return str(self.state.kernel["arch"])

    def _cross_compile(self) -> str:
        """COSS_COMPILE argument for kernel builds"""

        assert self.state.kernel is not None
        return str(self.state.kernel["cross_compile"])

    def configure(self, kernel_name, cross_compiler=None):
        """configure kernel build
        
        # Arguments
        kernel_name: name of the kernel config (see Metadata)
        cross_compiler: cross_compiler object to use
        """

        if cross_compiler is None:
            cross_compiler = self.board.compiler()

        super(KernelBuilder, self).configure(cross_compiler=cross_compiler)

        self._mkbuilddir()

        self.state.kernel = {}
        self.state.kernel["arch"] = (
            cross_compiler.machine.value
            if cross_compiler.machine.value != "aarch64"
            else "arm64"
        )
        self.state.kernel["kernel_name"] = kernel_name
        self.state.kernel["cross_compile"] = os.path.join(
            cross_compiler.bin_path, cross_compiler.prefix
        )

        kernel_desc = self._kernel_desc()
        self.state.srcdir = self.builddir / kernel_desc.kernel_srcdir
        self.state.prefix = Path("/")

        self._save_state()

        with self.context.cd(str(self.builddir)):
            srcdir = self.srcdir

            if not Path(srcdir).exists():
                self.context.run("cp {} .".format(kernel_desc.kernel_source))
                kernel_archive = (
                    self.builddir / Path(kernel_desc.kernel_source).name
                )
                untar(kernel_archive, self.builddir)

            with self.context.cd(str(srcdir)):

                self.context.run(
                    "cp {} .config".format(kernel_desc.kernel_config)
                )

                self.context.run(
                    "make ARCH={0} CROSS_COMPILE={1} olddefconfig".format(
                        self._arch(), self._cross_compile()
                    )
                )

                config_builder = KernelConfigBuilder(self.board, cross_compiler)
                config_fragment = srcdir / ".config_fragment"
                with config_fragment.open("w") as fragment:
                    fragment_str = config_builder.predefined_config_fragment(
                        kernel_name
                    )
                    print(fragment_str)
                    fragment.write(fragment_str)

                with self.context.prefix(
                    "export ARCH={0} && export CROSS_COMPILE={1}".format(
                        self._arch(), self._cross_compile()
                    )
                ):
                    self.context.run(
                        "./scripts/kconfig/merge_config.sh .config .config_fragment"
                    )

                self.context.run(
                    "cp .config {}".format(kernel_desc.kernel_config)
                )

    def build(self):
        """Build kernel"""

        self._mkbuilddir()
        kernel_desc = self._kernel_desc()

        build_path = Path(self.builddir)
        install_path = build_path / "install"
        boot_path = install_path / "boot"
        self.context.run("rm -rf {}".format(install_path))

        with self.context.cd(str(self.builddir)):
            srcdir = kernel_desc.kernel_srcdir
            with self.context.cd(str(srcdir)):

                self.logger.info("Trying build")

                self.context.run(
                    "make -j {2} ARCH={0} CROSS_COMPILE={1} all".format(
                        self._arch(),
                        self._cross_compile(),
                        self._num_build_cpus(),
                    )
                )

                self.logger.info("Trying dtbs install target")

                dtb_install_result = self.context.run(
                    "make dtbs_install ARCH={0} CROSS_COMPILE={1} INSTALL_DTBS_PATH={2}/lib/firmware/{3}/device-tree/".format(
                        self._arch(),
                        self._cross_compile(),
                        str(install_path),
                        str(kernel_desc.version),
                    ),
                    warn=True,
                )

                if dtb_install_result.return_code != 0:
                    self.logger.warning(
                        "Could not install kernel dtbs this command is only available for newer kernels"
                    )

                self.context.run(
                    "rm -f {}".format(
                        str(
                            install_path
                            / "lib"
                            / "modules"
                            / kernel_desc.version
                            / "build"
                        )
                    )
                )

                header_install_result = self.context.run(
                    "make headers_install ARCH={0} CROSS_COMPILE={1} INSTALL_HDR_PATH={2}".format(
                        self._arch(),
                        self._cross_compile(),
                        str(
                            install_path
                            / "lib"
                            / "modules"
                            / kernel_desc.version
                            / "build"
                        ),
                    ),
                    warn=True,
                )
                if header_install_result.return_code != 0:
                    self.logger.warning("Could not install kernel headers")

                self.context.run(
                    "make modules_install ARCH={0} CROSS_COMPILE={1} INSTALL_MOD_PATH={2}".format(
                        self._arch(), self._cross_compile(), str(install_path)
                    )
                )

            kernel_image = build_path / kernel_desc.image.build_path
            kernel_dest = (
                install_path / kernel_desc.image.deploy_path.relative_to("/")
            )

            kernel_dest.parent.mkdir(parents=True, exist_ok=True)
            self.context.run(
                "cp {0} {1}".format(str(kernel_image), str(kernel_dest))
            )

            if kernel_desc.uboot:
                build_ubimage(
                    self.context,
                    kernel_desc.uboot,
                    self._arch(),
                    build_path,
                    boot_path,
                    kernel_image,
                )

    def install(self):
        """Install kernel locally and save build directory snapshot
        """
        kernel_desc = self._kernel_desc()
        kernel_data = self._kernel_data()
        with self.context.cd(str(self.builddir)):
            kernel_dir = kernel_data.shared_data_dir
            with self.context.cd("install"):
                kernel_package = kernel_data.deploy_package_name

                self.context.run("tar czf {0} boot lib".format(kernel_package))

            kernel_top_dir = kernel_desc.kernel_srcdir.parts[0]
            kernel_build_cache = kernel_data.build_cache_name

            self.context.run(
                "tar cJf  {} {}".format(kernel_build_cache, kernel_top_dir)
            )
            self.context.run(
                "cp {} {}".format(
                    kernel_build_cache, kernel_data.build_cache_path
                )
            )

    def deploy(self):
        logging.warn("Deployment for kernels is currently not provided")



import logging
import os.path
from pathlib import Path

from invoke import Exit, task

from ..utils.network import rsync


@task
def run(c, board, command, cwd=""):  # pragma: no cover
    """Run command remotely

    -b/--board: target board id
    -c/--command: command to run
    --cwd: working directory of the command
    """

    bh = c.board(board)

    if not cwd:
        cwd = bh.model.rundir

    with bh.connect() as con:
        con.run("mkdir -p {}".format(str(cwd)))
        with con.cd(str(cwd)):
            con.run(command)


@task
def put(c, board, file, remote_path=""):  # pragma: no cover
    """Put file on the board

    -b/--board: target board id
    -f/--file: local file
    -r/--remote: remote file path (default is board specific rundir)
    """

    bh = c.board(board)
    with bh.connect() as con:
        if not remote_path:
            remote_path = bh.model.rundir
            remote_file = remote_path / Path(file).name

        else:
            remote_file = Path(remote_path)

        logging.info(
            "Putting %s to %s:%s", str(file), str(board), str(remote_file)
        )
        con.run("mkdir -p {}".format(str(remote_file.parent)))
        con.put(str(file), str(remote_file))


@task
def get(c, board, remote, local=""):  # pragma: no cover
    """Get file from board

    -b/--board: target board id
    -r/--remote: remote file path
    -l/--local:  local folder or filename (default is current working directory)
    """

    bh = c.board(board)
    with bh.connect() as con:
        if local:
            local_path = Path(local)
            if local_path.is_dir():
                local_path.mkdir(parents=True, exist_ok=True)
            else:
                local_path.parent.mkdir(parents=True, exist_ok=True)

        con.get(remote=str(remote), local=str(local))


@task
def lock(c, board, lease_time="1h", timeout=""):  # pragma: no cover
    """Lock board

    -b/--board: target board id
    -t/--timeout: timeout for the lock (deprecated use lease-time)
    -l/--lease-time: The board is kept locked for at least lease_time
         Lease time can be given as an int representing seconds or with a suffix
         d for days, h for hours, m for minutes, s for seconds
    """
    board = c.board(board)
    if timeout != "":
        logging.warning(
            "-t/--timeout has been deprecated please use -l/--lease-time instead"
        )
        lease_time = timeout
    board.lock(lease_time=lease_time)


@task
def unlock(c, board):  # pragma: no cover
    """Unlock board

    -b/--board: target board id
    """
    board = c.board(board)
    board.unlock()


@task
def is_locked(c, board):  # pragma: no cover
    """Return 0 if board is locked by other user

    -b/--board: target board id
    """
    board = c.board(board)
    if board.is_locked():
        logging.info("Board is locked by other user")
        raise Exit(code=0)

    logging.info("Board  is available")
    raise Exit(code=1)


@task
def has_lock(c, board):  # pragma: no cover
    """Return 0 if board is locked by current user

    -b/--board: target board id
    """
    board = c.board(board)
    if board.has_lock():
        logging.info("We have the board lock")
        raise Exit(code=0)

    logging.info("We do not have the board lock")
    raise Exit(code=1)


@task
def trylock(c, board, lease_time="1h", timeout=""):
    """Try to lock the board.

    Returns immediately, and returns 0 if lock has been acquired.

    -b/--board: target board name
    -l/--lease-time: duration of the lock
    """
    board = c.board()

    if timeout != "":
        logging.warning(
            "-t/--timeout has been deprecated please use -l/--lease-time instead"
        )
        lease_time = timeout

    if board.trylock(lease_time=lease_time):
        raise Exit(code=0)
    raise Exit(code=1)


@task
def reboot(c, board, wait=False):  # pragma: no cover
    """Reboot  board

    -b/--board: target board id
    -w/--wait block until the board is reachable via ssh again
    """

    board = c.board(board)
    board.reboot(wait)


@task
def reset(c, board, wait=False):  # pragma: no cover
    """Hard reset board

    -b/--board: target board id
    -w/--wait: block until the board is reachable again
    """

    board = c.board(board)
    board.reset(wait)


@task
def install(c, board, package):  # pragma: no cover
    """Install a package on the board

    -b/--board: target board id
    -p/--package: package
    """

    board = c.board(board)

    # TODO: Support other distributions as needed
    if board.os.distribution not in ["debian", "ubuntu"]:
        logging.error(
            "Currently package installation only supports Ubuntu or Debian based systems this board is {}".format(
                board.os.distribution
            )
        )
        raise Exit(code=-1)

    apt = "DEBIAN_FRONTEND=noninteractive sudo apt-get install -y {0}"
    with board.connect() as con:
        con.run(apt.format(package))

    raise Exit(code=0)


@task
def shell(c, board):  # pragma: no cover
    """Start a remote shell

    -b/--board: target board id
    """

    board = c.board(board)

    with board.connect() as con:
        con.run("$SHELL", pty=True)
        print("Finished")


@task
def board_ids(c, filter=""):  # pragma: no cover
    """returns list of board_ids suitable for usage in shell scripts

    -f/--filter: filter expression for boards

            Filter expression is prepended with 'lambda board:  ' and then evaluated as a python function
            board is an object of class Board, only returns board_ids if filter expression is true

            Examples:

            board.machine == 'zynqberry' to only run on zynqberry boards

            board.trylock() to only iterate over boards that are currently unlocked, and lock the boards while iterating
    """

    if filter:
        filter = "lambda board: " + filter
        filter = eval(filter)
    else:
        filter = lambda board: True

    for board in (board for board in c.boards() if filter(board)):
        print(board.id)


@task
def rsync_to(c, board, source, target="", delete=False):  # pragma: no cover
    """rsync a folder to the target board by default the


    -b/--board: target board id
    -s/--source: source folder/file
    -t/--target: target folder/file default is configured rundir on the board
    -d/--delete: delete files from target that do not exist in the source
    """
    board = c.board(board)

    if not target:
        target = str(board.rundir) + "/"
    elif target[0] != "/":
        target = os.path.join(board.rundir, target)

    with board.connect() as con:
        rsync(con, source=source, target=target, delete=delete)


@task
def kexec(
    c, board, kernel_id="", append="", commandline="", wait=False
):  # pragma: no cover
    """Start the Linux kernel using kexec

    -b/--board: target board id
    -k/--kernel-id: target kernel id
    -a/--append: Append the given string to the commandline
    -w/--wait: wait until board is reachable via ssh again
    """

    board = c.board(board)

    board.kexec(kernel_id, append, commandline, wait=wait)



import datetime
import logging
import time
from contextlib import AbstractContextManager, contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Any, List, Union

from fabric import Connection

from automate.model.board import SSHConnectionModel

from .builder import BaseBuilder, CMakeBuilder, KernelBuilder, MakefileBuilder
from .compiler import CrossCompiler
from .locks import DatabaseLockManager, SimpleLockManager
from .model import BoardModel, CompilerModel
from .model.common import Toolchain
from .utils.kernel import KernelData
from .utils.network import connect

if TYPE_CHECKING:
    import automate.context


class _ConnectionContextManager(AbstractContextManager):
    """Context manager for board connections to allow reuse of board connections"""

    def __init__(self, board, type: str = "ssh", timeout: int = 30):
        self.board = board
        self.timeout = timeout
        self.type = type
        self.nested = (
            board._connection is not None and board._connection.is_connected
        )
        self.locking_thread = None
        if not self.nested:
            self.locking_thread = board.lock_manager.keep_lock(board.name)

        self._connect()

    def _connect(self):
        board = self.board
        if self.type != "ssh":
            raise Exception("Currently only ssh connections are supported")
        if board._connection is not None and board._connection.is_connected:
            return board._connection

        for connection in board.model.connections:
            if isinstance(connection, SSHConnectionModel):
                host = connection.host
                user = connection.username
                port = connection.port

                extra_gateway_connection = None
                extra_gateway_config = self.board.context.config.automate.get(
                    "extra_gateway", None
                )
                if extra_gateway_config:
                    extra_gateway_connection = connect(
                        extra_gateway_config["host"],
                        extra_gateway_config["user"],
                        extra_gateway_config.get("port", 22),
                        passwd_allowed=True,
                    )

                gateway_connection = None
                if board.model.gateway:
                    gw_host = board.model.gateway.host
                    gw_user = board.model.gateway.username
                    gw_port = board.model.gateway.port

                    gateway_connection = connect(
                        gw_host,
                        gw_user,
                        gw_port,
                        identity=board.identity,
                        timeout=self.timeout,
                        gateway=extra_gateway_connection,
                    )

                c = connect(
                    host,
                    user,
                    port,
                    identity=board.identity,
                    gateway=gateway_connection,
                    timeout=self.timeout,
                )

                c.open()

                board._connection = c
                return c

        raise Exception(
            "Could not get ssh connection for {}".format(board.model.name)
        )

    def __enter__(self):
        return self.board._connection

    def __exit__(self, *exc_details):

        if self.locking_thread is not None:
            self.locking_thread.stop()
            self.locking_thread = None

        if not self.nested:
            c = self.board._connection
            self.board.connection = None
            c.__exit__(*exc_details)

    def __del__(self):
        if self.locking_thread is not None:
            self.locking_thread.stop()
            self.locking_thread = None

    def __getattr__(self, attr: str) -> Any:
        """proxy model properties if they are not shadowed by an own property"""
        return getattr(self.board._connection, attr)


class Board(object):
    """ Automation Class for script based interaction with boards:

    Provides access to:
       - board data
       - board automation: upload, execution, reboot, reset, ...
       - board specific cross compilers
    """

    def __init__(
        self,
        context: "automate.context.AutomateContext",
        board: BoardModel,
        compilers: List[CompilerModel],
        identity: Union[Path, str],
    ) -> None:
        self.logger = logging.getLogger(__name__)
        self.context = context
        self.model = board
        self.compiler_models = compilers
        self.identity = Path(identity).absolute()
        if context.database is not None:
            self.lock_manager: Union[
                DatabaseLockManager, SimpleLockManager
            ] = DatabaseLockManager(context.database)
        else:
            self.lock_manager = SimpleLockManager(
                Path(context.config.automate.boardroot) / "locks.db"
            )
        self._connection = None

    @property
    def id(self) -> str:
        # FIXME: Remove and replace with name
        return self.model.name

    @contextmanager
    def lock_ctx(self, lease_time: str = "1h"):
        if not self.has_lock():
            try:
                yield self.lock(lease_time=lease_time)
            finally:
                self.unlock()
        else:
            # Return a do nothing context manager
            try:
                yield None
            finally:
                pass

    def lock(self, lease_time: str = "1h"):
        self.lock_manager.lock(self, lease_time)

    def has_lock(self) -> bool:
        return self.lock_manager.has_lock(self)

    def unlock(self):
        return self.lock_manager.unlock(self)

    def trylock(self, lease_time: str = "1h") -> bool:
        return self.lock_manager.trylock(self, lease_time)

    def is_locked(self) -> bool:
        return self.lock_manager.is_locked(self)

    def lock_lease_time(self) -> datetime.timedelta:
        return self.lock_manager.lease_time(self)

    def lock_holder(self) -> str:
        return self.lock_manager.lock_holder(self)

    def compiler(
        self,
        compiler_name: str = "",
        toolchain: Union[Toolchain, str] = Toolchain.GCC,
    ) -> CrossCompiler:
        """ Build a Cross Compiler Object for this board 

        By default uses newest gcc compiler available in metadata.

        # Arguments
           compiler_name: use a specifc compiler id
           toolchain:  use newest configured compiler

        # Returns
           Object of class #automate.compiler.CrossCompiler configured to run builds for this board
        """

        if isinstance(toolchain, str):
            toolchain = Toolchain(toolchain)

        sorted_models = reversed(
            sorted(self.compiler_models, key=lambda x: x.version)
        )
        for compiler_model in sorted_models:
            if compiler_name != "":
                if compiler_name == compiler_model.name:
                    cc = CrossCompiler(self.context, compiler_model, self)
                    if cc.valid:
                        return cc
            else:
                cc = CrossCompiler(self.context, compiler_model, self)
                if cc.toolchain == toolchain and cc.valid:
                    return cc

        raise Exception(
            "Could not get compatible compiler with: name: '{}' and toolchain: '{}'".format(
                compiler_name, toolchain.value
            )
        )

    def compilers(
        self, toolchain: Union[Toolchain, None, str] = None
    ) -> List[CrossCompiler]:
        """ Build list of cross compiler objects configured for this board

        # Arguments
        toolchain: Only build cross compilers from the given toolchain

        # Returns
        List of configured cross compilers
        """

        if isinstance(toolchain, str):
            if toolchain == "":
                toolchain = None
            else:
                toolchain = Toolchain(toolchain)

        res = []
        for model in self.compiler_models:
            cc = CrossCompiler(self.context, model, self)
            if (cc.toolchain == toolchain) or (toolchain is None):
                if cc.valid:
                    res.append(cc)
        return res

    def connect(self, type: str = "ssh", timeout: int = 30) -> Connection:
        """
        Return a fabric.Connection to the board.

        # Arguments
        type: connection type currently only "ssh" is supportted
        timeout: timeout unitl connection should be established
        
        # Returns
        
        /fabric.Connection/ to the board

        """

        if self.is_locked():
            raise Exception("Can not connect to locked board")

        return _ConnectionContextManager(self, timeout=timeout, type=type)

    def reboot(self, wait=True) -> Union[Connection, None]:
        """ Starts a new connection to the device and initiates a reboot

        # Arguments
        wait: If wait is true tries to start a new connection, 
              waits until connecting succeeds, and returns a new connection.
        # Returns
        If wait was given a new connection is Returned
        """

        self.logger.info("Rebooting board {}".format(self.name))

        with self.connect() as connection:
            connection.run("sudo shutdown -r now & exit")

        self.logger.info("Reboot initiated!")

        if wait:
            return self.wait_for_connection()

        return None

    def wait_for_connection(self) -> Connection:
        """Wait until a successful ssh connection to the board can be established

        # Returns
        A new fabric.Connection object

        """

        self.logger.info("waiting for reconnection")
        connected = False
        while not connected:
            try:
                connection = self.connect()
                return connection
            except Exception:
                self.logger.info("Waiting for reconnection")
                time.sleep(3)

    def reset(self, wait=True) -> Union[Connection, None]:
        """Hard-Reset the board

        TODO: Currently not implemented

        # Arguments
        wait: if true wait until the board is connectible again

        # Returns
        If wait was true a new Connection object
        """

        if self.model.reset:
            gateway_connection = None
            extra_gateway_config = self.context.config.automate.get(
                "extra_gateway", None
            )
            if extra_gateway_config:
                extra_gateway_connection = connect(
                    extra_gateway_config["host"],
                    extra_gateway_config["user"],
                    extra_gateway_config.get("port", 22),
                    passwd_allowed=True,
                )

            if self.model.gateway:
                gw_host = self.model.gateway.host
                gw_user = self.model.gateway.username
                gw_port = self.model.gateway.port

                gateway_connection = connect(
                    gw_host,
                    gw_user,
                    gw_port,
                    identity=self.identity,
                    gateway=extra_gateway_connection,
                )
            if gateway_connection is not None:
                for command in self.model.reset:
                    gateway_connection.run(command, shell=True)

                gateway_connection.close()
        else:
            self.logger.warning(
                "True resets are currently not implemented, for board %s trying a reboot instead",
                self.name,
            )
            try:
                self.reboot(wait=False)
            except BaseException as e:
                self.logger.error(
                    "Reboot has been unsuccessful with exception {}".format(
                        str(e)
                    )
                )
                self.logger.error("Please reboot device manually")

        if wait:
            return self.wait_for_connection()

        return None

    def homedir(self) -> Path:
        """ Return the home directory of the connected user 
        
        # Returns

        pathlib.Path: home directory
        """
        with self.connect() as con:
            result = con.run("echo $HOME", hide=True)
            return Path(result.stdout.strip())

    def kexec(
        self, kernel_name="", append="", commandline="", wait=True
    ) -> Union[Connection, None]:
        """ Start a board kernel using kexec 

        # Arguments
        kernel_name: unique name of the kernel to boot
        append: string of addition kernel commandline flags
        commandline: completely new kernel commandline
        wait: block unitl board is reachable via ssh again and reconnect
        
        # Returns
        if wait was given a new fabric.Connection is returned
        """
        kernel_config = None

        for config in self.os.kernels:
            if kernel_name and kernel_name == config.name:
                kernel_config = config
                break
            elif not kernel_name and config.default:
                kernel_config = config
                break

        if kernel_config is None:
            raise Exception("Could not find kernel config")

        if kernel_config.image is None:
            raise Exception("Could not find the kernel image entry")

        image = str(kernel_config.image.deploy_path)
        if not commandline:
            commandline = kernel_config.commandline
        commandline = commandline + " " + append

        with self.connect() as con:
            with self.lock_ctx():
                cmd = "sudo kexec -l {} --command-line='{}'".format(
                    image, commandline
                )

                self.logger.info("Running {}".format(cmd))
                con.run(cmd)

                cmd = (
                    "nohup bash -c 'sleep 1; sudo kexec -e' > /tmp/nohup.out &"
                )
                self.logger.info("Running {}".format(cmd))
                con.run(cmd)
                self.logger.info("Kexec executed")
                time.sleep(2.0)

                if wait:
                    self.logger.info("Waiting for reconnection")
                    return self.wait_for_connection()

        return None

    def kernel_data(self, name: str) -> Union[KernelData, None]:
        """ Information about the installed kernels 

        # Arguments
        name: kernel name for which information should be returned
        
        # Returns
        KernelData object for the kernel configration
        
        """

        for kernel_desc in self.os.kernels:
            if kernel_desc.name == name:
                return KernelData(self, kernel_desc)
        return None

    def builder(self, typ, builddir: Union[Path, str] = "") -> BaseBuilder:
        """ Return a builder object for this board

        # Arguments
        typ: Type of the buildsystem to use choices are cmake, kernel, make
        
        
        # Returns
        configured builder object
        """

        if typ == "cmake":
            return CMakeBuilder(self.context, self, builddir)
        elif typ == "kernel":
            return KernelBuilder(self.context, self, builddir)
        elif typ == "make":
            return MakefileBuilder(self.context, self, builddir)

        raise Exception("Could not find builder {}".format(typ))

    def __getattr__(self, attr: str) -> Any:
        """proxy model properties if they are not shadowed by an own property"""
        return getattr(self.model, attr)
